from enum import Enum


class ProcessDtoType(str, Enum):
    GOLDEN = "GOLDEN"
    IMPORT = "IMPORT"
    LIVE = "LIVE"
    MASTER = "MASTER"
    NORMAL = "NORMAL"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class RoleDtoName(str, Enum):
    ADMIN = "ADMIN"
    API = "API"
    MANAGER = "MANAGER"
    MONITOR = "MONITOR"
    SUPER = "SUPER"
    USER = "USER"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from qubicon.api.types import UNSET, Unset

T = TypeVar("T", bound="AuthenticationTokenDto")


@_attrs_define
class AuthenticationTokenDto:
    """
    Attributes:
        token (Union[Unset, str]):
        token_type (Union[Unset, str]):
        expires_in (Union[Unset, int]):
        refresh_token (Union[Unset, str]):
        expiration_time (Union[Unset, int]):
    """

    token: Union[Unset, str] = UNSET
    token_type: Union[Unset, str] = UNSET
    expires_in: Union[Unset, int] = UNSET
    refresh_token: Union[Unset, str] = UNSET
    expiration_time: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        token = self.token

        token_type = self.token_type

        expires_in = self.expires_in

        refresh_token = self.refresh_token

        expiration_time = self.expiration_time

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if token is not UNSET:
            field_dict["token"] = token
        if token_type is not UNSET:
            field_dict["tokenType"] = token_type
        if expires_in is not UNSET:
            field_dict["expiresIn"] = expires_in
        if refresh_token is not UNSET:
            field_dict["refreshToken"] = refresh_token
        if expiration_time is not UNSET:
            field_dict["expirationTime"] = expiration_time

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        token = d.pop("token", UNSET)

        token_type = d.pop("tokenType", UNSET)

        expires_in = d.pop("expiresIn", UNSET)

        refresh_token = d.pop("refreshToken", UNSET)

        expiration_time = d.pop("expirationTime", UNSET)

        authentication_token_dto = cls(
            token=token,
            token_type=token_type,
            expires_in=expires_in,
            refresh_token=refresh_token,
            expiration_time=expiration_time,
        )

        authentication_token_dto.additional_properties = d
        return authentication_token_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union

import httpx

from .. import errors
from ..client import AuthenticatedClient, Client
from ...models.get_offline_equipment_statuses_item import GetOfflineEquipmentStatusesItem
from qubicon.api.types import UNSET, Response, Unset


def _get_kwargs(
    *,
    search: Union[Unset, str] = UNSET,
    sort: Union[Unset, List[str]] = UNSET,
    statuses: Union[Unset, List[GetOfflineEquipmentStatusesItem]] = UNSET,
    gmp: Union[Unset, bool] = UNSET,
    page: Union[Unset, int] = UNSET,
    size: Union[Unset, int] = UNSET,
) -> Dict[str, Any]:
    params: Dict[str, Any] = {}

    params["search"] = search

    json_sort: Union[Unset, List[str]] = UNSET
    if not isinstance(sort, Unset):
        json_sort = sort

    params["sort"] = json_sort

    json_statuses: Union[Unset, List[str]] = UNSET
    if not isinstance(statuses, Unset):
        json_statuses = []
        for statuses_item_data in statuses:
            statuses_item = statuses_item_data.value
            json_statuses.append(statuses_item)

    params["statuses"] = json_statuses

    params["gmp"] = gmp

    params["page"] = page

    params["size"] = size

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: Dict[str, Any] = {
        "method": "get",
        "url": "/public-api/offline-equipment",
        "params": params,
    }

    return _kwargs


def _parse_response(*, client: Union[AuthenticatedClient, Client], response: httpx.Response) -> Optional[Any]:
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: Union[AuthenticatedClient, Client], response: httpx.Response) -> Response[Any]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    search: Union[Unset, str] = UNSET,
    sort: Union[Unset, List[str]] = UNSET,
    statuses: Union[Unset, List[GetOfflineEquipmentStatusesItem]] = UNSET,
    gmp: Union[Unset, bool] = UNSET,
    page: Union[Unset, int] = UNSET,
    size: Union[Unset, int] = UNSET,
) -> Response[Any]:
    """
    Args:
        search (Union[Unset, str]):
        sort (Union[Unset, List[str]]):
        statuses (Union[Unset, List[GetOfflineEquipmentStatusesItem]]):
        gmp (Union[Unset, bool]):
        page (Union[Unset, int]):
        size (Union[Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Any]
    """

    kwargs = _get_kwargs(
        search=search,
        sort=sort,
        statuses=statuses,
        gmp=gmp,
        page=page,
        size=size,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    search: Union[Unset, str] = UNSET,
    sort: Union[Unset, List[str]] = UNSET,
    statuses: Union[Unset, List[GetOfflineEquipmentStatusesItem]] = UNSET,
    gmp: Union[Unset, bool] = UNSET,
    page: Union[Unset, int] = UNSET,
    size: Union[Unset, int] = UNSET,
) -> Response[Any]:
    """
    Args:
        search (Union[Unset, str]):
        sort (Union[Unset, List[str]]):
        statuses (Union[Unset, List[GetOfflineEquipmentStatusesItem]]):
        gmp (Union[Unset, bool]):
        page (Union[Unset, int]):
        size (Union[Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Any]
    """

    kwargs = _get_kwargs(
        search=search,
        sort=sort,
        statuses=statuses,
        gmp=gmp,
        page=page,
        size=size,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from qubicon.api.types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.xlsx_parsing_error_model_throwable import XlsxParsingErrorModelThrowable


T = TypeVar("T", bound="XlsxParsingErrorModel")


@_attrs_define
class XlsxParsingErrorModel:
    """
    Attributes:
        sheet_name (Union[Unset, str]):
        row_number (Union[Unset, int]):
        col_number (Union[Unset, int]):
        error_message (Union[Unset, str]):
        args (Union[Unset, List[str]]):
        throwable (Union[Unset, XlsxParsingErrorModelThrowable]):
    """

    sheet_name: Union[Unset, str] = UNSET
    row_number: Union[Unset, int] = UNSET
    col_number: Union[Unset, int] = UNSET
    error_message: Union[Unset, str] = UNSET
    args: Union[Unset, List[str]] = UNSET
    throwable: Union[Unset, "XlsxParsingErrorModelThrowable"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        sheet_name = self.sheet_name

        row_number = self.row_number

        col_number = self.col_number

        error_message = self.error_message

        args: Union[Unset, List[str]] = UNSET
        if not isinstance(self.args, Unset):
            args = self.args

        throwable: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.throwable, Unset):
            throwable = self.throwable.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if sheet_name is not UNSET:
            field_dict["sheetName"] = sheet_name
        if row_number is not UNSET:
            field_dict["rowNumber"] = row_number
        if col_number is not UNSET:
            field_dict["colNumber"] = col_number
        if error_message is not UNSET:
            field_dict["errorMessage"] = error_message
        if args is not UNSET:
            field_dict["args"] = args
        if throwable is not UNSET:
            field_dict["throwable"] = throwable

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.xlsx_parsing_error_model_throwable import XlsxParsingErrorModelThrowable

        d = src_dict.copy()
        sheet_name = d.pop("sheetName", UNSET)

        row_number = d.pop("rowNumber", UNSET)

        col_number = d.pop("colNumber", UNSET)

        error_message = d.pop("errorMessage", UNSET)

        args = cast(List[str], d.pop("args", UNSET))

        _throwable = d.pop("throwable", UNSET)
        throwable: Union[Unset, XlsxParsingErrorModelThrowable]
        if isinstance(_throwable, Unset):
            throwable = UNSET
        else:
            throwable = XlsxParsingErrorModelThrowable.from_dict(_throwable)

        xlsx_parsing_error_model = cls(
            sheet_name=sheet_name,
            row_number=row_number,
            col_number=col_number,
            error_message=error_message,
            args=args,
            throwable=throwable,
        )

        xlsx_parsing_error_model.additional_properties = d
        return xlsx_parsing_error_model

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from qubicon.api.types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.authentication_token_dto import AuthenticationTokenDto


T = TypeVar("T", bound="AuthenticationDto")


@_attrs_define
class AuthenticationDto:
    """
    Attributes:
        normal (Union[Unset, AuthenticationTokenDto]):
        monitor (Union[Unset, AuthenticationTokenDto]):
    """

    normal: Union[Unset, "AuthenticationTokenDto"] = UNSET
    monitor: Union[Unset, "AuthenticationTokenDto"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        normal: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.normal, Unset):
            normal = self.normal.to_dict()

        monitor: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.monitor, Unset):
            monitor = self.monitor.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if normal is not UNSET:
            field_dict["normal"] = normal
        if monitor is not UNSET:
            field_dict["monitor"] = monitor

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.authentication_token_dto import AuthenticationTokenDto

        d = src_dict.copy()
        _normal = d.pop("normal", UNSET)
        normal: Union[Unset, AuthenticationTokenDto]
        if isinstance(_normal, Unset):
            normal = UNSET
        else:
            normal = AuthenticationTokenDto.from_dict(_normal)

        _monitor = d.pop("monitor", UNSET)
        monitor: Union[Unset, AuthenticationTokenDto]
        if isinstance(_monitor, Unset):
            monitor = UNSET
        else:
            monitor = AuthenticationTokenDto.from_dict(_monitor)

        authentication_dto = cls(
            normal=normal,
            monitor=monitor,
        )

        authentication_dto.additional_properties = d
        return authentication_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.online_equipment_node_variable_dto_equipment_type import OnlineEquipmentNodeVariableDtoEquipmentType
from ..models.online_equipment_node_variable_dto_node_opc_ua_type import OnlineEquipmentNodeVariableDtoNodeOpcUaType
from ..models.online_equipment_node_variable_dto_signal_type import OnlineEquipmentNodeVariableDtoSignalType
from qubicon.api.types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.physical_quantity_unit_dto import PhysicalQuantityUnitDto
    from ..models.publish_schema_dto import PublishSchemaDto
    from ..models.sensor_type_dto import SensorTypeDto


T = TypeVar("T", bound="OnlineEquipmentNodeVariableDto")


@_attrs_define
class OnlineEquipmentNodeVariableDto:
    """
    Attributes:
        id (int):
        opc_ua_server_id (Union[Unset, int]):
        equipment_id (Union[Unset, int]):
        gmp_enabled (Union[Unset, bool]):
        opc_ua_server_name (Union[Unset, str]):
        equipment_name (Union[Unset, str]):
        browse_name (Union[Unset, str]):
        display_name (Union[Unset, str]):
        description (Union[Unset, str]):
        node_opc_ua_id (Union[Unset, str]):
        tag_name (Union[Unset, str]):
        qubicon_name (Union[Unset, str]):
        equipment_type (Union[Unset, OnlineEquipmentNodeVariableDtoEquipmentType]):
        node_opc_ua_type (Union[Unset, OnlineEquipmentNodeVariableDtoNodeOpcUaType]):
        mapped (Union[Unset, bool]):
        monitored (Union[Unset, bool]):
        health_status (Union[Unset, bool]):
        health_last_check_time (Union[Unset, int]):
        auto_start_reading (Union[Unset, bool]):
        opc_ua_data_type (Union[Unset, str]):
        compatible_data_type (Union[Unset, str]):
        physical_quantity_unit (Union[Unset, PhysicalQuantityUnitDto]):
        sensor_type (Union[Unset, SensorTypeDto]):
        signal_type (Union[Unset, OnlineEquipmentNodeVariableDtoSignalType]):
        actual_state_value (Union[Unset, str]):
        actual_state_change_time (Union[Unset, int]):
        min_value (Union[Unset, float]):
        max_value (Union[Unset, float]):
        publish_schema (Union[Unset, PublishSchemaDto]):
        single_value (Union[Unset, bool]):
    """

    id: int
    opc_ua_server_id: Union[Unset, int] = UNSET
    equipment_id: Union[Unset, int] = UNSET
    gmp_enabled: Union[Unset, bool] = UNSET
    opc_ua_server_name: Union[Unset, str] = UNSET
    equipment_name: Union[Unset, str] = UNSET
    browse_name: Union[Unset, str] = UNSET
    display_name: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    node_opc_ua_id: Union[Unset, str] = UNSET
    tag_name: Union[Unset, str] = UNSET
    qubicon_name: Union[Unset, str] = UNSET
    equipment_type: Union[Unset, OnlineEquipmentNodeVariableDtoEquipmentType] = UNSET
    node_opc_ua_type: Union[Unset, OnlineEquipmentNodeVariableDtoNodeOpcUaType] = UNSET
    mapped: Union[Unset, bool] = UNSET
    monitored: Union[Unset, bool] = UNSET
    health_status: Union[Unset, bool] = UNSET
    health_last_check_time: Union[Unset, int] = UNSET
    auto_start_reading: Union[Unset, bool] = UNSET
    opc_ua_data_type: Union[Unset, str] = UNSET
    compatible_data_type: Union[Unset, str] = UNSET
    physical_quantity_unit: Union[Unset, "PhysicalQuantityUnitDto"] = UNSET
    sensor_type: Union[Unset, "SensorTypeDto"] = UNSET
    signal_type: Union[Unset, OnlineEquipmentNodeVariableDtoSignalType] = UNSET
    actual_state_value: Union[Unset, str] = UNSET
    actual_state_change_time: Union[Unset, int] = UNSET
    min_value: Union[Unset, float] = UNSET
    max_value: Union[Unset, float] = UNSET
    publish_schema: Union[Unset, "PublishSchemaDto"] = UNSET
    single_value: Union[Unset, bool] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id

        opc_ua_server_id = self.opc_ua_server_id

        equipment_id = self.equipment_id

        gmp_enabled = self.gmp_enabled

        opc_ua_server_name = self.opc_ua_server_name

        equipment_name = self.equipment_name

        browse_name = self.browse_name

        display_name = self.display_name

        description = self.description

        node_opc_ua_id = self.node_opc_ua_id

        tag_name = self.tag_name

        qubicon_name = self.qubicon_name

        equipment_type: Union[Unset, str] = UNSET
        if not isinstance(self.equipment_type, Unset):
            equipment_type = self.equipment_type.value

        node_opc_ua_type: Union[Unset, str] = UNSET
        if not isinstance(self.node_opc_ua_type, Unset):
            node_opc_ua_type = self.node_opc_ua_type.value

        mapped = self.mapped

        monitored = self.monitored

        health_status = self.health_status

        health_last_check_time = self.health_last_check_time

        auto_start_reading = self.auto_start_reading

        opc_ua_data_type = self.opc_ua_data_type

        compatible_data_type = self.compatible_data_type

        physical_quantity_unit: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.physical_quantity_unit, Unset):
            physical_quantity_unit = self.physical_quantity_unit.to_dict()

        sensor_type: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.sensor_type, Unset):
            sensor_type = self.sensor_type.to_dict()

        signal_type: Union[Unset, str] = UNSET
        if not isinstance(self.signal_type, Unset):
            signal_type = self.signal_type.value

        actual_state_value = self.actual_state_value

        actual_state_change_time = self.actual_state_change_time

        min_value = self.min_value

        max_value = self.max_value

        publish_schema: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.publish_schema, Unset):
            publish_schema = self.publish_schema.to_dict()

        single_value = self.single_value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
            }
        )
        if opc_ua_server_id is not UNSET:
            field_dict["opcUaServerId"] = opc_ua_server_id
        if equipment_id is not UNSET:
            field_dict["equipmentId"] = equipment_id
        if gmp_enabled is not UNSET:
            field_dict["gmpEnabled"] = gmp_enabled
        if opc_ua_server_name is not UNSET:
            field_dict["opcUaServerName"] = opc_ua_server_name
        if equipment_name is not UNSET:
            field_dict["equipmentName"] = equipment_name
        if browse_name is not UNSET:
            field_dict["browseName"] = browse_name
        if display_name is not UNSET:
            field_dict["displayName"] = display_name
        if description is not UNSET:
            field_dict["description"] = description
        if node_opc_ua_id is not UNSET:
            field_dict["nodeOpcUaId"] = node_opc_ua_id
        if tag_name is not UNSET:
            field_dict["tagName"] = tag_name
        if qubicon_name is not UNSET:
            field_dict["qubiconName"] = qubicon_name
        if equipment_type is not UNSET:
            field_dict["equipmentType"] = equipment_type
        if node_opc_ua_type is not UNSET:
            field_dict["nodeOpcUaType"] = node_opc_ua_type
        if mapped is not UNSET:
            field_dict["mapped"] = mapped
        if monitored is not UNSET:
            field_dict["monitored"] = monitored
        if health_status is not UNSET:
            field_dict["healthStatus"] = health_status
        if health_last_check_time is not UNSET:
            field_dict["healthLastCheckTime"] = health_last_check_time
        if auto_start_reading is not UNSET:
            field_dict["autoStartReading"] = auto_start_reading
        if opc_ua_data_type is not UNSET:
            field_dict["opcUaDataType"] = opc_ua_data_type
        if compatible_data_type is not UNSET:
            field_dict["compatibleDataType"] = compatible_data_type
        if physical_quantity_unit is not UNSET:
            field_dict["physicalQuantityUnit"] = physical_quantity_unit
        if sensor_type is not UNSET:
            field_dict["sensorType"] = sensor_type
        if signal_type is not UNSET:
            field_dict["signalType"] = signal_type
        if actual_state_value is not UNSET:
            field_dict["actualStateValue"] = actual_state_value
        if actual_state_change_time is not UNSET:
            field_dict["actualStateChangeTime"] = actual_state_change_time
        if min_value is not UNSET:
            field_dict["minValue"] = min_value
        if max_value is not UNSET:
            field_dict["maxValue"] = max_value
        if publish_schema is not UNSET:
            field_dict["publishSchema"] = publish_schema
        if single_value is not UNSET:
            field_dict["singleValue"] = single_value

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.physical_quantity_unit_dto import PhysicalQuantityUnitDto
        from ..models.publish_schema_dto import PublishSchemaDto
        from ..models.sensor_type_dto import SensorTypeDto

        d = src_dict.copy()
        id = d.pop("id")

        opc_ua_server_id = d.pop("opcUaServerId", UNSET)

        equipment_id = d.pop("equipmentId", UNSET)

        gmp_enabled = d.pop("gmpEnabled", UNSET)

        opc_ua_server_name = d.pop("opcUaServerName", UNSET)

        equipment_name = d.pop("equipmentName", UNSET)

        browse_name = d.pop("browseName", UNSET)

        display_name = d.pop("displayName", UNSET)

        description = d.pop("description", UNSET)

        node_opc_ua_id = d.pop("nodeOpcUaId", UNSET)

        tag_name = d.pop("tagName", UNSET)

        qubicon_name = d.pop("qubiconName", UNSET)

        _equipment_type = d.pop("equipmentType", UNSET)
        equipment_type: Union[Unset, OnlineEquipmentNodeVariableDtoEquipmentType]
        if isinstance(_equipment_type, Unset):
            equipment_type = UNSET
        else:
            equipment_type = OnlineEquipmentNodeVariableDtoEquipmentType(_equipment_type)

        _node_opc_ua_type = d.pop("nodeOpcUaType", UNSET)
        node_opc_ua_type: Union[Unset, OnlineEquipmentNodeVariableDtoNodeOpcUaType]
        if isinstance(_node_opc_ua_type, Unset):
            node_opc_ua_type = UNSET
        else:
            node_opc_ua_type = OnlineEquipmentNodeVariableDtoNodeOpcUaType(_node_opc_ua_type)

        mapped = d.pop("mapped", UNSET)

        monitored = d.pop("monitored", UNSET)

        health_status = d.pop("healthStatus", UNSET)

        health_last_check_time = d.pop("healthLastCheckTime", UNSET)

        auto_start_reading = d.pop("autoStartReading", UNSET)

        opc_ua_data_type = d.pop("opcUaDataType", UNSET)

        compatible_data_type = d.pop("compatibleDataType", UNSET)

        _physical_quantity_unit = d.pop("physicalQuantityUnit", UNSET)
        physical_quantity_unit: Union[Unset, PhysicalQuantityUnitDto]
        if isinstance(_physical_quantity_unit, Unset):
            physical_quantity_unit = UNSET
        else:
            physical_quantity_unit = PhysicalQuantityUnitDto.from_dict(_physical_quantity_unit)

        _sensor_type = d.pop("sensorType", UNSET)
        sensor_type: Union[Unset, SensorTypeDto]
        if isinstance(_sensor_type, Unset):
            sensor_type = UNSET
        else:
            sensor_type = SensorTypeDto.from_dict(_sensor_type)

        _signal_type = d.pop("signalType", UNSET)
        signal_type: Union[Unset, OnlineEquipmentNodeVariableDtoSignalType]
        if isinstance(_signal_type, Unset):
            signal_type = UNSET
        else:
            signal_type = OnlineEquipmentNodeVariableDtoSignalType(_signal_type)

        actual_state_value = d.pop("actualStateValue", UNSET)

        actual_state_change_time = d.pop("actualStateChangeTime", UNSET)

        min_value = d.pop("minValue", UNSET)

        max_value = d.pop("maxValue", UNSET)

        _publish_schema = d.pop("publishSchema", UNSET)
        publish_schema: Union[Unset, PublishSchemaDto]
        if isinstance(_publish_schema, Unset):
            publish_schema = UNSET
        else:
            publish_schema = PublishSchemaDto.from_dict(_publish_schema)

        single_value = d.pop("singleValue", UNSET)

        online_equipment_node_variable_dto = cls(
            id=id,
            opc_ua_server_id=opc_ua_server_id,
            equipment_id=equipment_id,
            gmp_enabled=gmp_enabled,
            opc_ua_server_name=opc_ua_server_name,
            equipment_name=equipment_name,
            browse_name=browse_name,
            display_name=display_name,
            description=description,
            node_opc_ua_id=node_opc_ua_id,
            tag_name=tag_name,
            qubicon_name=qubicon_name,
            equipment_type=equipment_type,
            node_opc_ua_type=node_opc_ua_type,
            mapped=mapped,
            monitored=monitored,
            health_status=health_status,
            health_last_check_time=health_last_check_time,
            auto_start_reading=auto_start_reading,
            opc_ua_data_type=opc_ua_data_type,
            compatible_data_type=compatible_data_type,
            physical_quantity_unit=physical_quantity_unit,
            sensor_type=sensor_type,
            signal_type=signal_type,
            actual_state_value=actual_state_value,
            actual_state_change_time=actual_state_change_time,
            min_value=min_value,
            max_value=max_value,
            publish_schema=publish_schema,
            single_value=single_value,
        )

        online_equipment_node_variable_dto.additional_properties = d
        return online_equipment_node_variable_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.physical_quantity_unit_dto_status import PhysicalQuantityUnitDtoStatus
from qubicon.api.types import UNSET, Unset

T = TypeVar("T", bound="PhysicalQuantityUnitDto")


@_attrs_define
class PhysicalQuantityUnitDto:
    """
    Attributes:
        unit (str):
        name (str):
        id (Union[Unset, int]):
        physical_quantity_id (Union[Unset, int]):
        status (Union[Unset, PhysicalQuantityUnitDtoStatus]):
    """

    unit: str
    name: str
    id: Union[Unset, int] = UNSET
    physical_quantity_id: Union[Unset, int] = UNSET
    status: Union[Unset, PhysicalQuantityUnitDtoStatus] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        unit = self.unit

        name = self.name

        id = self.id

        physical_quantity_id = self.physical_quantity_id

        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "unit": unit,
                "name": name,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if physical_quantity_id is not UNSET:
            field_dict["physicalQuantityId"] = physical_quantity_id
        if status is not UNSET:
            field_dict["status"] = status

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        unit = d.pop("unit")

        name = d.pop("name")

        id = d.pop("id", UNSET)

        physical_quantity_id = d.pop("physicalQuantityId", UNSET)

        _status = d.pop("status", UNSET)
        status: Union[Unset, PhysicalQuantityUnitDtoStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = PhysicalQuantityUnitDtoStatus(_status)

        physical_quantity_unit_dto = cls(
            unit=unit,
            name=name,
            id=id,
            physical_quantity_id=physical_quantity_id,
            status=status,
        )

        physical_quantity_unit_dto.additional_properties = d
        return physical_quantity_unit_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from qubicon.api.types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item_annotations_item import (
        XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemAnnotationsItem,
    )
    from ..models.xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item_declared_annotations_item import (
        XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemDeclaredAnnotationsItem,
    )


T = TypeVar("T", bound="XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItem")


@_attrs_define
class XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItem:
    """
    Attributes:
        name (Union[Unset, str]):
        specification_title (Union[Unset, str]):
        specification_version (Union[Unset, str]):
        specification_vendor (Union[Unset, str]):
        implementation_title (Union[Unset, str]):
        implementation_version (Union[Unset, str]):
        implementation_vendor (Union[Unset, str]):
        sealed (Union[Unset, bool]):
        annotations (Union[Unset,
            List['XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemAnnotationsItem']]):
        declared_annotations (Union[Unset, List['XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPack
            agesItemDeclaredAnnotationsItem']]):
    """

    name: Union[Unset, str] = UNSET
    specification_title: Union[Unset, str] = UNSET
    specification_version: Union[Unset, str] = UNSET
    specification_vendor: Union[Unset, str] = UNSET
    implementation_title: Union[Unset, str] = UNSET
    implementation_version: Union[Unset, str] = UNSET
    implementation_vendor: Union[Unset, str] = UNSET
    sealed: Union[Unset, bool] = UNSET
    annotations: Union[
        Unset, List["XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemAnnotationsItem"]
    ] = UNSET
    declared_annotations: Union[
        Unset,
        List["XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemDeclaredAnnotationsItem"],
    ] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        specification_title = self.specification_title

        specification_version = self.specification_version

        specification_vendor = self.specification_vendor

        implementation_title = self.implementation_title

        implementation_version = self.implementation_version

        implementation_vendor = self.implementation_vendor

        sealed = self.sealed

        annotations: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.annotations, Unset):
            annotations = []
            for annotations_item_data in self.annotations:
                annotations_item = annotations_item_data.to_dict()
                annotations.append(annotations_item)

        declared_annotations: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.declared_annotations, Unset):
            declared_annotations = []
            for declared_annotations_item_data in self.declared_annotations:
                declared_annotations_item = declared_annotations_item_data.to_dict()
                declared_annotations.append(declared_annotations_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if name is not UNSET:
            field_dict["name"] = name
        if specification_title is not UNSET:
            field_dict["specificationTitle"] = specification_title
        if specification_version is not UNSET:
            field_dict["specificationVersion"] = specification_version
        if specification_vendor is not UNSET:
            field_dict["specificationVendor"] = specification_vendor
        if implementation_title is not UNSET:
            field_dict["implementationTitle"] = implementation_title
        if implementation_version is not UNSET:
            field_dict["implementationVersion"] = implementation_version
        if implementation_vendor is not UNSET:
            field_dict["implementationVendor"] = implementation_vendor
        if sealed is not UNSET:
            field_dict["sealed"] = sealed
        if annotations is not UNSET:
            field_dict["annotations"] = annotations
        if declared_annotations is not UNSET:
            field_dict["declaredAnnotations"] = declared_annotations

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item_annotations_item import (
            XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemAnnotationsItem,
        )
        from ..models.xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item_declared_annotations_item import (
            XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemDeclaredAnnotationsItem,
        )

        d = src_dict.copy()
        name = d.pop("name", UNSET)

        specification_title = d.pop("specificationTitle", UNSET)

        specification_version = d.pop("specificationVersion", UNSET)

        specification_vendor = d.pop("specificationVendor", UNSET)

        implementation_title = d.pop("implementationTitle", UNSET)

        implementation_version = d.pop("implementationVersion", UNSET)

        implementation_vendor = d.pop("implementationVendor", UNSET)

        sealed = d.pop("sealed", UNSET)

        annotations = []
        _annotations = d.pop("annotations", UNSET)
        for annotations_item_data in _annotations or []:
            annotations_item = XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemAnnotationsItem.from_dict(
                annotations_item_data
            )

            annotations.append(annotations_item)

        declared_annotations = []
        _declared_annotations = d.pop("declaredAnnotations", UNSET)
        for declared_annotations_item_data in _declared_annotations or []:
            declared_annotations_item = XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemDeclaredAnnotationsItem.from_dict(
                declared_annotations_item_data
            )

            declared_annotations.append(declared_annotations_item)

        xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item = cls(
            name=name,
            specification_title=specification_title,
            specification_version=specification_version,
            specification_vendor=specification_vendor,
            implementation_title=implementation_title,
            implementation_version=implementation_version,
            implementation_vendor=implementation_vendor,
            sealed=sealed,
            annotations=annotations,
            declared_annotations=declared_annotations,
        )

        xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item.additional_properties = d
        return xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ProcessPhaseDtoStatus(str, Enum):
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    NOT_READY = "NOT_READY"
    READY = "READY"
    RUNNING = "RUNNING"

    def __str__(self) -> str:
        return str(self.value)



# CUSTOM SDK LOGIC (ONTOP OF AUTO-GENERATED OPENAPI CLIENT)

from qubicon.api.public_api_controller import (
    create_computable_model,
    create_physical_quantity,
    get_channels,
    get_equipment_variable_nodes,
    get_job_artifact,
    get_job,
    get_list_of_computable_models,
    get_multiplex_chart_data_channels,
    get_offline_equipment,
    get_online_equipments,
    get_physical_quantities,
    get_processes,
    start_export_process_data,
    stream_events,
    login,
    refresh,
)

from qubicon.api.client import AuthenticatedClient

from qubicon.models import (
    UserLoginDto,
    UserLoginDtoLoginType,
    RefreshTokenDto,
    PhysicalQuantityDto,
    ExternalPythonComputableModelDto,
    JythonComputableModelDto,
    PublicProcessExportRequestDto,
    MultiplexChartDataChannelsRequestDto,
    ComputableModelInputDto,
    ComputableModelOutputDto,
    PhysicalQuantityUnitDto,
    AbstractComputableModelDto,
    AbstractComputableModelDtoEngineType,
    AbstractComputableModelDtoCalculationStyle,
    AbstractComputableModelDtoStatus,
    PhysicalQuantityUnitDtoStatus
)
from qubicon.api.types import UNSET
import pandas as pd
from http import HTTPStatus
from typing import Any, Dict, List, Optional
import json
import logging
from rapidfuzz import fuzz
import re

class QubiconCore:
    def __init__(self, client: AuthenticatedClient):
        """
        Keep the entire HTTP client here, which contains:
            - base_url
            - token (updated upon login)
            - any request config
        """
        self.client = client

    def login_user(self, username: str, password: str, login_type: str = "INTERNAL") -> Optional[str]:
        """Login user and return authentication token."""
        login_type_enum = UserLoginDtoLoginType(login_type)
        body = UserLoginDto(username=username, password=password, login_type=login_type_enum)
        print("Serialized Payload:", body.to_dict())

        try:
            response = login.sync_detailed(client=self.client, body=body)

            # Debugging: Log raw server response
            print(f"Login Response Status: {response.status_code}")
            print(f"Login Response Content: {response.content}")

            if response.status_code == HTTPStatus.OK:
                # Extract and parse JSON response
                parsed_response = json.loads(response.content.decode())
                print("Parsed Response:", parsed_response)  # Debugging

                # Safely retrieve the token
                token = parsed_response.get("normal", {}).get("token")
                if not token:
                    raise ValueError("Token not found in response.")

                # Update the client with the new token
                self.client.token = token

                print("Login successful! Token set in client.")
                return token
            else:
                print(f"Login failed: {response.content}")
                return None
        except Exception as e:
            print(f"Error logging in: {e}")
            return None

    def login_user1(self, username: str, password: str, login_type: str = "INTERNAL") -> Optional[str]:
        """Login user and return authentication token."""
        login_type_enum = UserLoginDtoLoginType(login_type)
        body = UserLoginDto(username=username, password=password, login_type=login_type_enum)
        try:
            response = login.sync_detailed(client=self.client, body=body)
            if response.status_code == HTTPStatus.OK:
                print("Login successful!")
                return response.parsed.get("token")
            else:
                print(f"Login failed: {response.content}")
                return None
        except Exception as e:
            print(f"Error logging in: {e}")
            return None
        
    def refresh_token(self, refresh_token: str) -> Optional[str]:
        """Refresh authentication token."""
        body = RefreshTokenDto(refresh_token=refresh_token)
        try:
            response = refresh.sync_detailed(client=self.client, body=body)
            if response.status_code == HTTPStatus.OK:
                print("Token refreshed successfully!")
                return response.parsed.get("accessToken")
            else:
                print(f"Token refresh failed: {response.content}")
                return None
        except Exception as e:
            print(f"Error refreshing token: {e}")
            return None

    def list_computable_models(
        self,
        statuses: list = UNSET,
        calculation_styles: list = UNSET,
        types: list = UNSET,
        output_physical_quantity_unit_ids: list = UNSET,
        output_physical_quantity_ids: list = UNSET,
        name: str = UNSET,
        deleted: bool = UNSET,
        sort: list = UNSET,
        page: int = UNSET,
        size: int = 1000,
    ):
        """
        Fetch all computable models with optional filters applied.

        Args:
            statuses (list, optional): Filter by statuses (e.g., ['RELEASED', 'DRAFT']).
            calculation_styles (list, optional): Filter by calculation styles.
            types (list, optional): Filter by model types.
            output_physical_quantity_unit_ids (list, optional): Filter by output physical quantity unit IDs.
            output_physical_quantity_ids (list, optional): Filter by output physical quantity IDs.
            name (str, optional): Filter by name.
            deleted (bool, optional): Filter by deletion status.
            sort (list, optional): Sort results (e.g., ['name,asc']).
            page (int, optional): Specify the page number for pagination.
            size (int, optional): Specify the number of results per page (default: 1000).

        Returns:
            List[Dict] or Dict: A list of computable models or an error message in JSON format.
        """
        try:
            # Fetch computable models using the OpenAPI function with filters applied
            response = get_list_of_computable_models.sync_detailed(
                client=self.client,
                statuses=statuses,
                calculation_styles=calculation_styles,
                types=types,
                output_physical_quantity_unit_ids=output_physical_quantity_unit_ids,
                output_physical_quantity_ids=output_physical_quantity_ids,
                name=name,
                deleted=deleted,
                sort=sort,
                page=page,
                size=size,
            )

            # Ensure the response status is OK
            if response.status_code == HTTPStatus.OK:
                # Parse the JSON content
                return json.loads(response.content.decode())
            else:
                logging.error(f"Failed to fetch computable models. HTTP Status: {response.status_code}")
                return {"error": f"Failed to fetch computable models. HTTP Status: {response.status_code}"}

        except Exception as e:
            logging.error(f"Error fetching computable models: {e}")
            return {"error": f"Error fetching computable models: {str(e)}"}

    def fetch_model_details(self, model_id: int):
        """
        Fetch details for a specific model by filtering from the list of models.

        Args:
            model_id (int): The ID of the model to fetch details for.

        Returns:
            Dict or None: The details of the model if found, or None if not found.
        """
        try:
            # Fetch all computable models
            response = get_list_of_computable_models.sync_detailed(client=self.client, size=1000)

            if response.status_code == HTTPStatus.OK:
                # Parse the JSON content
                parsed_response = json.loads(response.content.decode())

                # Check if 'content' exists and filter for the model with the given ID
                if "content" in parsed_response and isinstance(parsed_response["content"], list):
                    models = parsed_response["content"]

                    # Search for the model with the matching ID
                    for model in models:
                        if model["id"] == model_id:
                            return model  # Return the specific model

                    # If the ID is not found, return None
                    logging.warning(f"Model with ID {model_id} not found.")
                    return None
                else:
                    logging.error("Unexpected response format: 'content' key not found or invalid.")
                    return None
            else:
                logging.error(f"Failed to fetch models. HTTP Status: {response.status_code}")
                return None

        except Exception as e:
            logging.error(f"Error fetching model details: {e}")
            return None

        except Exception as e:
            logging.error(f"Error fetching model details: {e}")
            return None

    def list_physical_quantities(self):
        """
        List all physical quantities and return them as JSON.

        Returns:
            List[Dict] or Dict: A list of physical quantities as dictionaries or an error message if the request fails.
        """
        try:
            # Fetch the physical quantities using the OpenAPI function
            response = get_physical_quantities.sync_detailed(client=self.client)

            # Ensure the response status is OK
            if response.status_code == HTTPStatus.OK:
                # Parse and return the JSON content
                return json.loads(response.content.decode())
            else:
                # Log and return an error message in JSON format
                logging.error(f"Failed to fetch physical quantities. HTTP Status: {response.status_code}")
                logging.error(f"Response: {response.content.decode()}")
                return {"error": f"Failed to fetch physical quantities. HTTP Status: {response.status_code}"}

        except Exception as e:
            # Log and return an error message in JSON format
            logging.error(f"Error fetching physical quantities: {e}")
            return {"error": f"Error fetching physical quantities: {str(e)}"}

    def list_processes(
        self,
        name: str = UNSET,
        deleted: bool = UNSET,
        start_date: int = UNSET,
        end_date: int = UNSET,
        statuses: list = UNSET,
        archived_or_will_be_archived: bool = UNSET,
        types: list = UNSET,
        ids: list = UNSET,
        lot_ids: list = UNSET,
        organism_ids: list = UNSET,
        kpi_model_ids: list = UNSET,
        material_ids: list = UNSET,
        stream_ids: list = UNSET,
        user_stream_ids: list = UNSET,
        online_equipment_ids: list = UNSET,
        offline_equipment_ids: list = UNSET,
        online_equipment_group_first_ids: list = UNSET,
        offline_equipment_group_first_ids: list = UNSET,
        recipe_ids: list = UNSET,
        organism_and_vial_tuples: list = UNSET,
        material_and_lot_tuples: list = UNSET,
        gmp: bool = UNSET,
        imported: bool = UNSET,
        group_ids: list = UNSET,
        master_recipe_ids: list = UNSET,
        part_of_experiment: bool = UNSET,
        sampling_values: list = UNSET,
        sort: list = UNSET,
        page: int = UNSET,
        size: int = 10000,
    ):
        """
        List all processes with optional filters applied.

        Args:
            name (str, optional): Filter by process name.
            deleted (bool, optional): Filter by deleted status.
            start_date (int, optional): Filter by processes starting on/after this date (Unix timestamp in ms).
            end_date (int, optional): Filter by processes ending on/before this date (Unix timestamp in ms).
            statuses (list, optional): Filter by process statuses (e.g., ['COMPLETED']).
            archived_or_will_be_archived (bool, optional): Filter by archived or to-be-archived processes.
            types (list, optional): Filter by process types (e.g., ['IMPORT']).
            ids (list, optional): Filter by specific process IDs.
            lot_ids (list, optional): Filter by specific lot IDs.
            organism_ids (list, optional): Filter by specific organism IDs.
            kpi_model_ids (list, optional): Filter by specific KPI model IDs.
            material_ids (list, optional): Filter by specific material IDs.
            stream_ids (list, optional): Filter by specific stream IDs.
            user_stream_ids (list, optional): Filter by specific user-defined stream IDs.
            online_equipment_ids (list, optional): Filter by specific online equipment IDs.
            offline_equipment_ids (list, optional): Filter by specific offline equipment IDs.
            online_equipment_group_first_ids (list, optional): Filter by first online equipment group IDs.
            offline_equipment_group_first_ids (list, optional): Filter by first offline equipment group IDs.
            recipe_ids (list, optional): Filter by specific recipe IDs.
            organism_and_vial_tuples (list, optional): Filter by tuples of organism and vial IDs.
            material_and_lot_tuples (list, optional): Filter by tuples of material and lot IDs.
            gmp (bool, optional): Filter by GMP-compliant processes.
            imported (bool, optional): Filter by imported processes.
            group_ids (list, optional): Filter by specific group IDs.
            master_recipe_ids (list, optional): Filter by master recipe IDs.
            part_of_experiment (bool, optional): Filter by whether processes are part of an experiment.
            sampling_values (list, optional): Filter by specific sampling values.
            sort (list, optional): Sort results (e.g., ['name,asc']).
            page (int, optional): Specify the page number for pagination.
            size (int, optional): Specify the number of results per page (default: 1000).

        Returns:
            List[Dict] or Dict: A list of processes or an error message in JSON format.
        """
        try:
            # Fetch processes using the OpenAPI function with filters applied
            response = get_processes.sync_detailed(
                client=self.client,
                name=name,
                deleted=deleted,
                start_date=start_date,
                end_date=end_date,
                statuses=statuses,
                archived_or_will_be_archived=archived_or_will_be_archived,
                types=types,
                ids=ids,
                lot_ids=lot_ids,
                organism_ids=organism_ids,
                kpi_model_ids=kpi_model_ids,
                material_ids=material_ids,
                stream_ids=stream_ids,
                user_stream_ids=user_stream_ids,
                online_equipment_ids=online_equipment_ids,
                offline_equipment_ids=offline_equipment_ids,
                online_equipment_group_first_ids=online_equipment_group_first_ids,
                offline_equipment_group_first_ids=offline_equipment_group_first_ids,
                recipe_ids=recipe_ids,
                organism_and_vial_tuples=organism_and_vial_tuples,
                material_and_lot_tuples=material_and_lot_tuples,
                gmp=gmp,
                imported=imported,
                group_ids=group_ids,
                master_recipe_ids=master_recipe_ids,
                part_of_experiment=part_of_experiment,
                sampling_values=sampling_values,
                sort=sort,
                page=page,
                size=size,
            )
            print(f"Response Status Code: {response.status_code}")


            # Ensure the response status is OK
            if response.status_code == HTTPStatus.OK:
                return json.loads(response.content.decode())
            else:
                logging.error(f"Failed to fetch processes. HTTP Status: {response.status_code}")
                return {"error": f"Failed to fetch processes. HTTP Status: {response.status_code}"}

        except Exception as e:
            logging.error(f"Error fetching processes: {e}")
            return {"error": f"Error fetching processes: {str(e)}"}
        
    def extract_channels(
        self,
        process_id: int,
        process_phase_id: int = UNSET,
        name: str = UNSET,
        node_types: list = UNSET,
        sensor_type_ids: list = UNSET,
        physical_quantity_unit_id: int = UNSET,
    ):
        """
        Extract channels for a given process ID with optional filters.

        Args:
            process_id (int): ID of the process to fetch channels for.
            process_phase_id (int, optional): Filter by specific process phase ID.
            name (str, optional): Filter by channel name.
            node_types (list, optional): Filter by node types (e.g., ['INPUT', 'OUTPUT']).
            sensor_type_ids (list, optional): Filter by specific sensor type IDs.
            physical_quantity_unit_id (int, optional): Filter by physical quantity unit ID.

        Returns:
            List[Dict] or Dict: A list of channels or an error message in JSON format.
        """
        try:
            # Call the relevant OpenAPI function with the provided arguments
            response = get_channels.sync_detailed(
                process_id=process_id,
                client=self.client,
                process_phase_id=process_phase_id,
                name=name,
                node_types=node_types,
                sensor_type_ids=sensor_type_ids,
                physical_quantity_unit_id=physical_quantity_unit_id,
            )

            # Check if the response status is OK
            if response.status_code == HTTPStatus.OK:
                print(json.dumps(json.loads(response.content.decode()), indent=4))
                return json.loads(response.content.decode())
            else:
                logging.error(f"Failed to fetch channels. HTTP Status: {response.status_code}")
                return {"error": f"Failed to fetch channels. HTTP Status: {response.status_code}"}

        except Exception as e:
            logging.error(f"Error fetching channels: {e}")
            return {"error": f"Error fetching channels: {str(e)}"}

    def extract_process_data(
        self,
        process_id: int,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        raw_mode: bool = False,
        export_rate: Optional[int] = None,
        at_sampling_timestamps_mode: bool = False,
        included_report: bool = False,
        start_event_id: Optional[int] = None,
        end_event_id: Optional[int] = None,
        output_format: str = "csv",
        output_file: str = "process_data.csv",
    ):
        """
        Extract process data for the specified process ID with filters and options.

        Args:
            process_id (int): ID of the process to extract data for.
            start_time (Optional[int]): Start time (Unix timestamp in ms).
            end_time (Optional[int]): End time (Unix timestamp in ms).
            raw_mode (bool): Enable raw mode (default: False).
            export_rate (Optional[int]): Granularity of the export in milliseconds.
            at_sampling_timestamps_mode (bool): Export at sampling timestamps (default: False).
            included_report (bool): Include report in the export (default: False).
            start_event_id (Optional[int]): ID of the start event.
            end_event_id (Optional[int]): ID of the end event.
            output_format (str): The desired output format ("csv" or "json").
            output_file (str): Path to the output file.

        Returns:
            Dict: The JSON response from the API or an error message.
        """
        try:
            # Create the DTO object with validated fields
            body = PublicProcessExportRequestDto(
                raw_mode=raw_mode,
                export_rate=export_rate,
                at_sampling_timestamps_mode=at_sampling_timestamps_mode,
                included_report=included_report,
                start_event_id=start_event_id,
                end_event_id=end_event_id,
                start_time=start_time,
                end_time=end_time,
            )

            # Call the OpenAPI function
            response = start_export_process_data.sync_detailed(
                process_id=process_id,
                client=self.client,
                body=body,
            )

            # Ensure the response status is OK
            if response.status_code == HTTPStatus.OK:
                # Parse the raw content
                parsed_response = json.loads(response.content.decode())
                print(json.dumps(parsed_response, indent=2))  # Debugging: Print the parsed response

                # Handle the data extraction based on format
                if "data" in parsed_response:  # Adjusted to check for 'data' key
                    data = parsed_response["data"]

                    if output_format == "csv":
                        # Convert to DataFrame and save as CSV
                        df = pd.DataFrame(data)
                        df.to_csv(output_file, index=False)
                        logging.info(f"Process data successfully exported to {output_file} in CSV format.")
                    elif output_format == "json":
                        # Save as JSON
                        with open(output_file, "w") as f:
                            json.dump(data, f, indent=4)
                        logging.info(f"Process data successfully exported to {output_file} in JSON format.")
                    else:
                        logging.error(f"Invalid output format: {output_format}")
                        return {"error": f"Invalid output format: {output_format}"}

                    return {"success": f"Process data exported to {output_file}"}
                else:
                    logging.error("Unexpected response format: 'data' key not found.")
                    return {"error": "Unexpected response format: 'data' key not found."}

            else:
                logging.error(f"Failed to export process data. HTTP Status: {response.status_code}")
                return {"error": f"Failed to export process data. HTTP Status: {response.status_code}"}

        except Exception as e:
            logging.error(f"Error exporting process data: {e}")
            return {"error": f"Error exporting process data: {str(e)}"}

    def stream_events(self, event_types: List[str], nonce: str):
        """Stream events from the API."""
        try:
            response = stream_events.sync(client=self.client, types=event_types, nonce=nonce)
            if response:
                for event in response:
                    print(f"Event: {event}")
            else:
                print("No events streamed.")
        except Exception as e:
            print(f"Error streaming events: {e}")

    def convert_model_to_importable_json(self, api_response):
        """
        Converts the API response of a model into an importable JSON structure.
        
        Args:
            api_response (dict): The API response from the GET model details endpoint.
            
        Returns:
            dict: The importable JSON structure.
        """
        if not api_response:
            logging.error("No valid model data to convert.")
            return None

        return {
            "engineType": api_response.get("engineType"),
            "kpiName": api_response.get("kpiName"),
            "abbr": api_response.get("abbr"),
            "calculationStyle": api_response.get("calculationStyle"),
            "status": "DRAFT",
            "description": api_response.get("description", ""),
            "script": api_response.get("script", ""),
            "inputs": [
                {
                    "name": input_item.get("name"),
                    "order": idx,
                    "physicalQuantityUnit": input_item.get("physicalQuantityUnit", {}),
                    "description": input_item.get("description", "")
                }
                for idx, input_item in enumerate(api_response.get("inputs", []))
            ],
            "outputs": [
                {
                    "name": output_item.get("name"),
                    "order": idx,
                    "physicalQuantityUnit": output_item.get("physicalQuantityUnit", {}),
                    "description": output_item.get("description", "")
                }
                for idx, output_item in enumerate(api_response.get("outputs", []))
            ]
        }

    def export_model_to_json(self, model_id: int, filename: str):
        """Export model details to a JSON file."""
        model_details = self.fetch_model_details(model_id)

        if not model_details:
            logging.error("Model details not found or could not be exported.")
            return
        
        importable_json = self.convert_model_to_importable_json(model_details)

        if not importable_json:
            logging.error("Failed to convert model to importable format.")
            return

        try:
            with open(filename, "w") as file:
                json.dump(model_details, file, indent=4)
            logging.info(f"Model details exported to '{filename}' successfully.")
        except Exception as e:
            logging.error(f"Error exporting model details: {e}")

    def create_computable_model(self, model_data: dict) -> Optional[Dict[str, Any]]:
        """
        Create a new computable model using the provided JSON data,
        ensuring we handle all Enums and sub-DTOs manually so that
        there's no 'str' object has no attribute 'value' error.
        """
        try:
            # ------------------------------------------------------
            # Step 1: Basic cleanup & mandatory checks
            # ------------------------------------------------------
            if "engineType" not in model_data:
                raise ValueError("Missing 'engineType' in model data.")

            # Remove fields that must NOT be submitted
            for ro_field in ["id", "creationDate", "updateDate"]:
                model_data.pop(ro_field, None)

            # If sensorType is None, remove it
            if model_data.get("sensorType") is None:
                model_data.pop("sensorType", None)

            # ------------------------------------------------------
            # Step 2: Handle physical quantities (existing logic)
            # ------------------------------------------------------
            model_data = self.handle_physical_quantities(model_data)
            if not model_data:
                raise ValueError("Failed to process physical quantities.")

            # ------------------------------------------------------
            # Step 3: Manually build all sub-dtos (inputs/outputs)
            # ------------------------------------------------------
            input_dto_list = []
            for inp in model_data.get("inputs", []):
                # Safely build PhysicalQuantityUnitDto if present
                pq_unit_dict = inp.get("physicalQuantityUnit", {})
                # Convert "status" if it exists
                raw_pq_status = pq_unit_dict.get("status", UNSET)
                if raw_pq_status not in [UNSET, None]:
                    raw_pq_status = PhysicalQuantityUnitDtoStatus(raw_pq_status)

                pq_unit_dto = PhysicalQuantityUnitDto(
                    unit=pq_unit_dict.get("unit", ""),
                    name=pq_unit_dict.get("name", ""),
                    id=pq_unit_dict.get("id", UNSET),
                    physical_quantity_id=pq_unit_dict.get("physicalQuantityId", UNSET),
                    status=raw_pq_status,
                )

                # Build the ComputableModelInputDto (note: no publishField here)
                input_dto = ComputableModelInputDto(
                    name=inp.get("name", ""),
                    physical_quantity_unit=pq_unit_dto,
                    id=inp.get("id", UNSET),
                    description=inp.get("description", UNSET),
                    order=inp.get("order", UNSET),
                )
                input_dto_list.append(input_dto)

            output_dto_list = []
            for outp in model_data.get("outputs", []):
                # If "publishField" is None, remove it so we don't break the constructor
                if outp.get("publishField") is None:
                    outp.pop("publishField", None)

                pq_unit_dict = outp.get("physicalQuantityUnit", {})
                # Convert "status" if it exists
                raw_pq_status = pq_unit_dict.get("status", UNSET)
                if raw_pq_status not in [UNSET, None]:
                    raw_pq_status = PhysicalQuantityUnitDtoStatus(raw_pq_status)

                pq_unit_dto = PhysicalQuantityUnitDto(
                    unit=pq_unit_dict.get("unit", ""),
                    name=pq_unit_dict.get("name", ""),
                    id=pq_unit_dict.get("id", UNSET),
                    physical_quantity_id=pq_unit_dict.get("physicalQuantityId", UNSET),
                    status=raw_pq_status,
                )

                # Build the ComputableModelOutputDto
                output_dto = ComputableModelOutputDto(
                    name=outp.get("name", ""),
                    physical_quantity_unit=pq_unit_dto,
                    id=outp.get("id", UNSET),
                    description=outp.get("description", UNSET),
                    order=outp.get("order", UNSET),
                    # If you want a publishField, pass it here; if not, do UNSET
                    publish_field=UNSET,
                )
                output_dto_list.append(output_dto)

            # ------------------------------------------------------
            # Step 4: Convert top-level strings into Enums, build final DTO
            # ------------------------------------------------------
            # engineType => AbstractComputableModelDtoEngineType
            engine_type_str = model_data["engineType"]  # e.g. "JYTHON" or "EXTERNAL_PYTHON"
            engine_type_enum = AbstractComputableModelDtoEngineType(engine_type_str)

            # calculationStyle => AbstractComputableModelDtoCalculationStyle
            calc_style_str = model_data.get("calculationStyle", "ONLINE")
            calc_style_enum = AbstractComputableModelDtoCalculationStyle(calc_style_str)

            # status => AbstractComputableModelDtoStatus
            status_str = model_data.get("status", "DRAFT")
            status_enum = AbstractComputableModelDtoStatus(status_str)

            # Choose which top-level class to build
            if engine_type_enum == AbstractComputableModelDtoEngineType.JYTHON:
                model_body = JythonComputableModelDto(
                    engine_type=engine_type_enum,
                    kpi_name=model_data.get("kpiName", ""),
                    abbr=model_data.get("abbr", ""),
                    calculation_style=calc_style_enum,
                    outputs=output_dto_list,
                    inputs=input_dto_list,  # pass your built inputs
                    status=status_enum,
                    description=model_data.get("description", UNSET),
                    script=model_data.get("script", UNSET),
                    # creation_date / update_date can remain UNSET
                )
            else:
                # EXTERNAL_PYTHON or other
                model_body = ExternalPythonComputableModelDto(
                    engine_type=engine_type_enum,
                    kpi_name=model_data.get("kpiName", ""),
                    abbr=model_data.get("abbr", ""),
                    calculation_style=calc_style_enum,
                    outputs=output_dto_list,
                    inputs=input_dto_list,
                    status=status_enum,
                    description=model_data.get("description", UNSET),
                    script=model_data.get("script", UNSET),
                )

            # ------------------------------------------------------
            # Step 5: Call the API
            # ------------------------------------------------------
            response = create_computable_model.sync_detailed(
                client=self.client,
                body=model_body,
            )

            if response.status_code == HTTPStatus.OK:
                parsed = json.loads(response.content.decode())
                logging.info("Successfully created computable model.")
                return parsed
            elif response.status_code == HTTPStatus.BAD_REQUEST:
                # If the server returns 400 with valid.unique.kpi.name, log a friendlier message
                if "valid.unique.kpi.name" in json.loads(response.content.decode()):
                    logging.error(
                        "Failed to create model because 'kpiName' is already in use. "
                        "Please choose a different kpiName."
                    )
            else:
                msg = (
                    f"Failed to create computable model. "
                    f"HTTP Status: {response.status_code}. "
                    f"Response: {response.content.decode()}"
                )
                logging.error(msg)
                return None

        except Exception as e:
            logging.error(f"Error creating computable model: {e}")
            return None

    def create_physical_quantity(self, pq_dict: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """
        Create a new physical quantity (or possibly retrieve it if the server
        returns 200 instead of 201). We rely on the server's logic to handle
        duplicates by name.

        Returns the parsed JSON on success, or None on error.
        """
        try:
            # Convert input dict to the generated DTO.
            # Make sure the keys in pq_dict match PhysicalQuantityDto’s fields (name, units, etc.).
            dto = PhysicalQuantityDto.from_dict(pq_dict)

            # Call the generated POST function
            response = create_physical_quantity.sync_detailed(
                client=self.client,
                body=dto,
            )

            # Some servers return 201 for “created”, some might return 200 if it already existed.
            if response.status_code in (HTTPStatus.CREATED, HTTPStatus.OK):
                # Manually parse the body into a dict
                return json.loads(response.content.decode())
            else:
                logging.error(
                    f"Failed to create physical quantity. "
                    f"HTTP Status: {response.status_code}. "
                    f"Response: {response.content.decode()}"
                )
                return None

        except Exception as e:
            logging.error(f"Error creating physical quantity: {e}")
            return None
    
    def check_existing_physical_quantity(self, pq_name: str) -> Optional[dict]:
        """
        Check if a physical quantity already exists in the system.

        Args:
            pq_name (str): Name of the physical quantity to check.

        Returns:
            dict or None: The physical quantity data if found, or None if not found.
        """
        try:
            response = get_physical_quantities.sync_detailed(client=self.client)
            if response.status_code in (HTTPStatus.OK, HTTPStatus.CREATED):
                parsed_response = json.loads(response.content.decode())
                for pq in parsed_response:
                    if self.fuzzy_match(pq['name'], pq_name):
                        return pq
                return None
            else:
                logging.error(f"Failed to fetch physical quantities. HTTP Status: {response.status_code}")
                return None
        except Exception as e:
            logging.error(f"Error fetching physical quantities: {e}")
            return None

    def handle_physical_quantities(self, model_data: dict) -> Optional[dict]:
        """
        Handle physical quantities and map them correctly for both inputs and outputs.

        Args:
            model_data (dict): Model data containing inputs and outputs with physical quantities.

        Returns:
            dict or None: Updated model data or None if mapping fails.
        """
        try:
            for var in model_data["inputs"] + model_data["outputs"]:
                pq_name = var["physicalQuantityUnit"]["name"]
                unit_name = var["physicalQuantityUnit"]["unit"]

                # Check if the physical quantity already exists
                existing_pq = self.check_existing_physical_quantity(pq_name)

                if existing_pq:
                    logging.info(f"Found existing physical quantity '{pq_name}' (ID: {existing_pq['id']}).")
                    var["physicalQuantityUnit"]["physicalQuantityId"] = existing_pq["id"]

                    # Check if the unit exists within the physical quantity
                    existing_unit = self.check_existing_unit(existing_pq, unit_name)

                    if existing_unit:
                        var["physicalQuantityUnit"]["unit"] = existing_unit["unit"]
                        var["physicalQuantityUnit"]["id"] = existing_unit["id"]
                    else:
                        updated_pq = self.add_unit_to_existing_physical_quantity(existing_pq["id"], pq_name, unit_name)
                        if updated_pq:
                            new_unit = next((u for u in updated_pq["units"] if u["unit"] == unit_name), None)
                            if new_unit:
                                var["physicalQuantityUnit"]["unit"] = new_unit["unit"]
                                var["physicalQuantityUnit"]["id"] = new_unit["id"]
                            else:
                                raise ValueError(f"Failed to retrieve the new unit '{unit_name}'.")
                else:
                    # Create a new physical quantity
                    new_pq = self.create_physical_quantity({"name": pq_name, "units": [{"unit": unit_name}]})
                    if new_pq:
                        var["physicalQuantityUnit"]["physicalQuantityId"] = new_pq["id"]
                        new_unit = new_pq["units"][0]
                        var["physicalQuantityUnit"]["unit"] = new_unit["unit"]
                        var["physicalQuantityUnit"]["id"] = new_unit["id"]
                    else:
                        raise ValueError(f"Failed to create physical quantity '{pq_name}'.")

            return model_data
        except Exception as e:
            logging.error(f"Error handling physical quantities: {e}")
            return None

    def check_existing_unit(self, physical_quantity: dict, unit_name: str) -> Optional[dict]:
        """
        Match a unit within a physical quantity.

        Args:
            physical_quantity (dict): The physical quantity data.
            unit_name (str): The name of the unit to match.

        Returns:
            dict or None: Matching unit or None if not found.
        """
        for unit in physical_quantity['units']:
            if self.fuzzy_match(unit['unit'], unit_name):
                logging.info(f"Unit match found: '{unit_name}' within physical quantity '{physical_quantity['name']}' (ID: {unit['id']}).")
                return unit
        logging.info(f"No unit match found for '{unit_name}' in physical quantity '{physical_quantity['name']}'.")
        return None

    def normalize_name(self, name: str) -> str:
        """Normalize physical quantity or unit names."""
        return re.sub(r'[^a-zA-Z0-9]', '', name).lower()

    def fuzzy_match(self, name1: str, name2: str, threshold: int = 93) -> bool:
        """Perform fuzzy matching between two names."""
        normalized_name1 = self.normalize_name(name1)
        normalized_name2 = self.normalize_name(name2)
        similarity = fuzz.token_sort_ratio(normalized_name1, normalized_name2)
        return similarity >= threshold

    def add_unit_to_existing_physical_quantity(self, pq_id: int, pq_name: str, unit_name: str) -> Optional[Dict[str, Any]]:
        """
        Work around the absence of an 'update' method by:
          1) Fetching the existing physical quantity (via check_existing_physical_quantity).
          2) Appending the new unit to the 'units' array.
          3) Calling create_physical_quantity(...) again with the updated data.
             This assumes the backend merges or updates based on PQ 'name'.

        Returns the updated PQ data on success or None on failure.
        """
        # 1) Get the existing physical quantity by name
        existing_pq = self.check_existing_physical_quantity(pq_name)
        if not existing_pq:
            logging.error(f"Physical quantity '{pq_name}' not found for updating.")
            return None

        logging.info(f"Adding unit '{unit_name}' to existing physical quantity '{pq_name}' (ID: {pq_id}).")

        # 2) Append new unit to the array
        #    existing_pq['units'] is a list of dicts like [{"unit": "L", "physicalQuantityId": X}, ...]
        units = existing_pq.get("units", [])
        units.append({"unit": unit_name, "physicalQuantityId": pq_id})

        # 3) Construct data that we will pass back into create_physical_quantity
        #    We typically keep name, status, and all units.
        updated_data = {
            "name": pq_name,
            "status": existing_pq.get("status", "IN_USE"),
            "units": units,
        }
        # NOTE: The "id" field might or might not break the re-POST—depends on backend rules.
        # Sometimes it's safest to omit "id" because the server may reject duplicates. 
        # But if your backend accepts an "id" on create, you can include it:
        #
        # updated_data["id"] = pq_id

        # Now call create_physical_quantity(...) again
        updated_pq = self.create_physical_quantity(updated_data)
        if updated_pq:
            logging.info(f"Unit '{unit_name}' added successfully to '{pq_name}'.")
            return updated_pq
        else:
            logging.error(f"Failed to add unit '{unit_name}' to physical quantity '{pq_name}'.")
            return None

    def list_recipes(
        self,
        page: int = 0,
        size: int = 400,
        type: str = "NORMAL",
        statuses: List[str] = ["DRAFT", "RELEASED", "GMP"],
        sort: List[str] = ["updateDate,desc"],
        search: str = "",
    ) -> Optional[Dict[str, Any]]:
        """
        List all recipes with optional query parameters.

        Args:
            page (int, optional): Page number. Default=0
            size (int, optional): Page size. Default=400
            type (str, optional): Filter by recipe type. Default='NORMAL'
            statuses (List[str], optional): Filter by statuses. Default=['DRAFT','RELEASED','GMP']
            sort (List[str], optional): Sorting fields. Default=['updateDate,desc']
            search (str, optional): Search string. Default=''

        Returns:
            Dict or None: A dict containing recipes or None if request fails.
        """
        base_url = self.client.base_url  # Or however you retrieve the base URL
        endpoint = f"{base_url}/api/control-recipes"
        params = {
            "page": page,
            "size": size,
            "type": type,
            "statuses": ",".join(statuses),
            "sort": sort,
            "search": search,
        }

        try:
            response = self.client.get_httpx_client().get(endpoint, params=params)
            if response.status_code == HTTPStatus.OK:
                return response.json()
            else:
                logging.error(
                    f"Failed to list recipes. HTTP {response.status_code}: {response.text}"
                )
                return None
        except Exception as e:
            logging.error(f"Error listing recipes: {e}")
            return None

    def get_recipe(self, recipe_id: int) -> Optional[Dict[str, Any]]:
        """
        Fetch a single recipe by its ID.

        Args:
            recipe_id (int): The unique ID of the recipe.

        Returns:
            Dict or None: A dict with recipe details, or None on failure.
        """
        base_url = self.client.base_url
        endpoint = f"{base_url}/api/control-recipes/{recipe_id}"
        params = {"deleted": "false"}  # from your example

        try:
            response = self.client.get_httpx_client().get(endpoint, params=params)
            if response.status_code == HTTPStatus.OK:
                return response.json()
            else:
                logging.error(
                    f"Failed to get recipe {recipe_id}. HTTP {response.status_code}: {response.text}"
                )
                return None
        except Exception as e:
            logging.error(f"Error getting recipe {recipe_id}: {e}")
            return None

    def create_process(
        self,
        name: str,
        recipe_id: int,
        description: str = "",
        process_type: str = "NORMAL",
        online_equipment_ids: Optional[List[int]] = None,
        offline_equipment_ids: Optional[List[int]] = None,
        process_materials: Optional[List[Any]] = None,
        process_organisms: Optional[List[Any]] = None,
        streams: Optional[List[Any]] = None,
        alert_users: Optional[List[Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Create a new process by POSTing to /api/processes.

        Args:
            name (str): The name of the new process.
            recipe_id (int): The recipe ID on which to base this process (aka processDefinitionId).
            description (str, optional): A short description of the process.
            process_type (str, optional): Type of process (usually 'NORMAL').
            online_equipment_ids (List[int], optional): Any associated online equipment IDs.
            offline_equipment_ids (List[int], optional): Any associated offline equipment IDs.
            process_materials (List, optional): If you want to pass any material references.
            process_organisms (List, optional): If you want to pass organism references.
            streams (List, optional): Any streams for the process.
            alert_users (List, optional): IDs or references for users to alert.

        Returns:
            Dict or None: The newly created process as a dict, or None on error.
        """
        base_url = self.client.base_url
        endpoint = f"{base_url}/api/processes"

        payload = {
            "name": name,
            "description": description,
            "type": process_type,
            "processDefinitionId": recipe_id,  # i.e. 'recipe' id
            "onlineEquipmentIds": online_equipment_ids or [],
            "offlineEquipmentIds": offline_equipment_ids or [],
            "processMaterials": process_materials or [],
            "processOrganisms": process_organisms or [],
            "streams": streams or [],
            "alertUsers": alert_users or [],
        }

        try:
            response = self.client.get_httpx_client().post(endpoint, json=payload)
            if 200 <= response.status_code < 300:
                return response.json()  # Return the newly created process
            else:
                logging.error(
                    f"Failed to create process. HTTP {response.status_code}: {response.text}"
                )
                return None
        except Exception as e:
            logging.error(f"Error creating process: {e}")
            return None

    def create_process_group(self, name: str, status: str = "RELEASED") -> Optional[Dict[str, Any]]:
        """
        Create a new process group by POSTing to /api/process-groups.

        Args:
            name (str): Name of the process group.
            status (str, optional): Group status. Typically 'RELEASED'.

        Returns:
            Dict or None: The newly created group or None on error.
        """
        base_url = self.client.base_url
        endpoint = f"{base_url}/api/process-groups"

        body = {
            "name": name,
            "status": status
        }

        try:
            response = self.client.get_httpx_client().post(endpoint, json=body)
            if 200 <= response.status_code < 300:
                return response.json()
            else:
                logging.error(
                    f"Failed to create process group. HTTP {response.status_code}: {response.text}"
                )
                return None
        except Exception as e:
            logging.error(f"Error creating process group: {e}")
            return None

    def add_processes_to_group(self, group_id: int, process_ids: List[int]) -> Optional[Dict[str, Any]]:
        """
        Add one or more processes to an existing group by PUTting the entire group object
        including the updated 'processes' array.

        Args:
            group_id (int): The existing process group ID.
            process_ids (List[int]): A list of process IDs to include in this group.

        Returns:
            Dict or None: Updated group data or None on error.
        """
        base_url = self.client.base_url
        endpoint = f"{base_url}/api/process-groups/{group_id}"

        # 1) Fetch the current group object
        try:
            current_group_resp = self.client.get_httpx_client().get(endpoint)
            if current_group_resp.status_code != HTTPStatus.OK:
                logging.error(
                    f"Failed to fetch group {group_id}. HTTP {current_group_resp.status_code}: {current_group_resp.text}"
                )
                return None
            group_data = current_group_resp.json()
        except Exception as e:
            logging.error(f"Error fetching group {group_id} before update: {e}")
            return None

        # 2) Update the 'processes' list with the new IDs
        existing_processes = group_data.get("processes", [])
        # each item typically looks like {"id": 2876, "name": "...", "status": "READY", ...}

        # Convert current IDs to a set for quick membership checks
        existing_ids = set(p["id"] for p in existing_processes if "id" in p)

        for pid in process_ids:
            if pid not in existing_ids:
                # Append minimal structure that the backend expects
                existing_processes.append({"id": pid, "status": "READY"})

        group_data["processes"] = existing_processes

        # 3) PUT the entire group object back
        try:
            put_response = self.client.get_httpx_client().put(endpoint, json=group_data)
            if 200 <= put_response.status_code < 300:
                return put_response.json()
            else:
                logging.error(
                    f"Failed to add processes to group {group_id}. "
                    f"HTTP {put_response.status_code}: {put_response.text}"
                )
                return None
        except Exception as e:
            logging.error(f"Error updating group {group_id}: {e}")
            return None

# Internal function (not for public use)
def _internal_function():
    pass

# Dynamically create __all__
__all__ = [name for name in globals() if not name.startswith("_")]


from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from qubicon.api.types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.online_equipment_dto import OnlineEquipmentDto
    from ..models.page_meta_data_dto import PageMetaDataDto


T = TypeVar("T", bound="PageDtoOnlineEquipmentDto")


@_attrs_define
class PageDtoOnlineEquipmentDto:
    """
    Attributes:
        content (Union[Unset, List['OnlineEquipmentDto']]):
        page (Union[Unset, PageMetaDataDto]):
    """

    content: Union[Unset, List["OnlineEquipmentDto"]] = UNSET
    page: Union[Unset, "PageMetaDataDto"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        content: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.content, Unset):
            content = []
            for content_item_data in self.content:
                content_item = content_item_data.to_dict()
                content.append(content_item)

        page: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.page, Unset):
            page = self.page.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if content is not UNSET:
            field_dict["content"] = content
        if page is not UNSET:
            field_dict["page"] = page

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.online_equipment_dto import OnlineEquipmentDto
        from ..models.page_meta_data_dto import PageMetaDataDto

        d = src_dict.copy()
        content = []
        _content = d.pop("content", UNSET)
        for content_item_data in _content or []:
            content_item = OnlineEquipmentDto.from_dict(content_item_data)

            content.append(content_item)

        _page = d.pop("page", UNSET)
        page: Union[Unset, PageMetaDataDto]
        if isinstance(_page, Unset):
            page = UNSET
        else:
            page = PageMetaDataDto.from_dict(_page)

        page_dto_online_equipment_dto = cls(
            content=content,
            page=page,
        )

        page_dto_online_equipment_dto.additional_properties = d
        return page_dto_online_equipment_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from qubicon.api.types import UNSET, Unset

T = TypeVar("T", bound="PublicProcessExportRequestDto")


@_attrs_define
class PublicProcessExportRequestDto:
    """
    Attributes:
        raw_mode (Union[Unset, bool]):
        export_rate (Union[Unset, int]):
        at_sampling_timestamps_mode (Union[Unset, bool]):
        included_report (Union[Unset, bool]):
        start_event_id (Union[Unset, int]):
        end_event_id (Union[Unset, int]):
        start_time (Union[Unset, int]):
        end_time (Union[Unset, int]):
    """

    raw_mode: Union[Unset, bool] = UNSET
    export_rate: Union[Unset, int] = UNSET
    at_sampling_timestamps_mode: Union[Unset, bool] = UNSET
    included_report: Union[Unset, bool] = UNSET
    start_event_id: Union[Unset, int] = UNSET
    end_event_id: Union[Unset, int] = UNSET
    start_time: Union[Unset, int] = UNSET
    end_time: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        raw_mode = self.raw_mode

        export_rate = self.export_rate

        at_sampling_timestamps_mode = self.at_sampling_timestamps_mode

        included_report = self.included_report

        start_event_id = self.start_event_id

        end_event_id = self.end_event_id

        start_time = self.start_time

        end_time = self.end_time

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if raw_mode is not UNSET:
            field_dict["rawMode"] = raw_mode
        if export_rate is not UNSET:
            field_dict["exportRate"] = export_rate
        if at_sampling_timestamps_mode is not UNSET:
            field_dict["atSamplingTimestampsMode"] = at_sampling_timestamps_mode
        if included_report is not UNSET:
            field_dict["includedReport"] = included_report
        if start_event_id is not UNSET:
            field_dict["startEventId"] = start_event_id
        if end_event_id is not UNSET:
            field_dict["endEventId"] = end_event_id
        if start_time is not UNSET:
            field_dict["startTime"] = start_time
        if end_time is not UNSET:
            field_dict["endTime"] = end_time

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        raw_mode = d.pop("rawMode", UNSET)

        export_rate = d.pop("exportRate", UNSET)

        at_sampling_timestamps_mode = d.pop("atSamplingTimestampsMode", UNSET)

        included_report = d.pop("includedReport", UNSET)

        start_event_id = d.pop("startEventId", UNSET)

        end_event_id = d.pop("endEventId", UNSET)

        start_time = d.pop("startTime", UNSET)

        end_time = d.pop("endTime", UNSET)

        public_process_export_request_dto = cls(
            raw_mode=raw_mode,
            export_rate=export_rate,
            at_sampling_timestamps_mode=at_sampling_timestamps_mode,
            included_report=included_report,
            start_event_id=start_event_id,
            end_event_id=end_event_id,
            start_time=start_time,
            end_time=end_time,
        )

        public_process_export_request_dto.additional_properties = d
        return public_process_export_request_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import argparse
import logging
import time
from qubicon.core import QubiconCore, AuthenticatedClient
from rich.console import Console, Group
from rich.table import Table
from rich.panel import Panel
from rich.prompt import Prompt, IntPrompt
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.text import Text
from rich.layout import Layout
from rich import print, box
from tabulate import tabulate
from datetime import datetime
from qubicon.api.types import UNSET
import json
from rich.prompt import Confirm
import pandas as pd

# Initialize a Rich console for improved output formatting
console = Console()

# Configure logging to output to both file and console
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
    handlers=[
        logging.FileHandler("qubicon_client.log"),
        logging.StreamHandler(),
    ],
)

# Constants for server options
SERVER_OPTIONS = {
    "1": "https://master.qub-lab.io/",
    "2": "https://release-3.5.qub-lab.io",
}

# Globals for maintaining the client and core state
client = None
qubicon_core = None

def choose_server() -> str:
    """Prompt the user to select a server or provide a custom URL."""
    console.print("[cyan]Choose a server to connect to:[/cyan]")
    console.print("1. ", SERVER_OPTIONS["1"])
    console.print("2. ", SERVER_OPTIONS["2"])
    choice = Prompt.ask("Enter 1, 2, or a custom URL", default="1")
    return SERVER_OPTIONS.get(choice, choice)

def setup_client(base_url: str, token: str):
    """Initialize the authenticated client and core logic."""
    global client, qubicon_core
    client = AuthenticatedClient(base_url=base_url, token=token)
    qubicon_core = QubiconCore(client)

def handle_login():
    """Handle user login and set up the client."""
    base_url = choose_server()
    username = Prompt.ask("Username")
    password = Prompt.ask("Password", password=True)

    console.print("[cyan]Attempting to log in...[/cyan]")
    global qubicon_core
    temp_client = AuthenticatedClient(base_url=base_url, token=None)
    temp_core = QubiconCore(temp_client)

    try:
        token = temp_core.login_user(username, password)
        if token:
            console.print("[green]Login successful![/green]")
            setup_client(base_url, token)
        else:
            console.print("[red]Login failed. Check your credentials and try again.[/red]")
    except Exception as e:
        console.print(f"[red]Login error: {e}[/red]")

def handle_login_testing():
    """Handle user login and set up the client."""
    base_url = "https://master.qub-lab.io/"
    username = "qubiconClient"
    password = "qubiconClient1!"

    console.print("[cyan]Attempting to log in...[/cyan]")
    temp_client = AuthenticatedClient(base_url=base_url, token=None)
    temp_core = QubiconCore(temp_client)

    try:
        token = temp_core.login_user(username, password)
        if token:
            console.print("[green]Login successful![/green]")
            setup_client(base_url, token)
        else:
            console.print("[red]Login failed. Check your credentials and try again.[/red]")
    except Exception as e:
        console.print(f"[red]Login error: {e}[/red]")

def ensure_logged_in():
    """Verify that the user is logged in before performing actions."""
    global client, qubicon_core
    if not client or not qubicon_core:
        console.print("[yellow]You are not logged in. Please log in first.[/yellow]")
        handle_login()

def display_menu():
    """Show a compact yet visually appealing menu."""
    menu_table = Table(
        box=box.ROUNDED,
        show_header=False,
        padding=(0, 1),
        min_width=60,
        border_style="bright_black"
    )
    menu_table.add_column("Option", justify="right", width=10, style="bold cyan")
    menu_table.add_column("Action", justify="left", style="dim")
    menu_table.title = "Qubicon CLI Menu"

    # Authentication
    menu_table.add_row("[bold]0[/]", "[white]Login[/]")
    menu_table.add_row("1", "Logout")
    menu_table.add_row("[bold]2[/]", "[white]Exit[/]")

    # Models
    menu_table.add_row("[dim] Models ───[/]", "", style="bright_white")
    menu_table.add_row("3", "List models")
    menu_table.add_row("4", "View model details")
    menu_table.add_row("5", "Create model")

    # Quantities
    menu_table.add_row("[dim] Quantities ───[/]", "", style="bright_white")
    menu_table.add_row("6", "List physical quantities")
    menu_table.add_row("7", "Create physical quantity")

    # Processes
    menu_table.add_row("[dim] Processes ───[/]", "", style="bright_white")
    menu_table.add_row("8", "List processes")
    menu_table.add_row("9", "Create process (from recipe)")
    menu_table.add_row("10", "List channels")

    # Data
    menu_table.add_row("[dim] Data ───[/]", "", style="bright_white")
    menu_table.add_row("11", "Export process data")
    menu_table.add_row("12", "Export model")

    # Recipes & Groups
    menu_table.add_row("[dim] Recipes & Groups ───[/]", "", style="bright_white")
    menu_table.add_row("13", "List recipes")
    menu_table.add_row("14", "View recipe details")
    menu_table.add_row("15", "Create group")
    menu_table.add_row("16", "Add processes to group")

    console.print(menu_table)

def interactive_mode():
    """Interactive loop with improved visual feedback."""
    while True:
        display_menu()
        choice = Prompt.ask(
            "[bold cyan]➤ Select operation[/]",
            choices=[str(i) for i in range(17)],
            show_choices=False
        )

        # Keep existing handler logic
        if choice == "0": handle_login()
        elif choice == "1": handle_logout()
        elif choice == "2":
            console.print("[bold green]Exiting. Goodbye![/]")
            break
        elif choice == "3": display_models()
        elif choice == "4": fetch_model_details()
        elif choice == "5": create_computable_model()
        elif choice == "6": list_physical_quantities()
        elif choice == "7": pass  # create_physical_quantity()
        elif choice == "8": list_processes()
        elif choice == "9": handle_create_process()
        elif choice == "10": fetch_process_channels()
        elif choice == "11": handle_process_data_extraction()
        elif choice == "12": export_model_in_importable_format()
        elif choice == "13": handle_list_recipes()
        elif choice == "14": handle_get_recipe()
        elif choice == "15": handle_create_process_group()
        elif choice == "16": handle_add_processes_to_group()
        else: console.print("[red]Invalid choice. Try again.[/]")

def handle_list_recipes():
    """CLI flow for listing recipes."""
    global qubicon_core
    if not qubicon_core:
        console.print("[red]Error: You must be logged in before listing recipes.[/red]")
        return

    page = IntPrompt.ask("Page (default=0)", default=0)
    size = IntPrompt.ask("Page size (default=400)", default=400)
    recipe_type = Prompt.ask("Recipe type (default=NORMAL)", default="NORMAL")
    statuses_str = Prompt.ask("Statuses (comma-separated, default='DRAFT,RELEASED,GMP')", default="DRAFT,RELEASED,GMP")
    statuses = [s.strip() for s in statuses_str.split(",")]
    sort_str = Prompt.ask("Sort fields (comma-separated, default='updateDate,desc')", default="updateDate,desc")
    sort = [s.strip() for s in sort_str.split(",")]
    search = Prompt.ask("Search string (default='')", default="")

    console.print("[cyan]Listing recipes...[/cyan]")
    result = qubicon_core.list_recipes(
        page=page,
        size=size,
        type=recipe_type,
        statuses=statuses,
        sort=sort,
        search=search,
    )
    if result is None:
        console.print("[red]Failed to list recipes.[/red]")
    else:
        console.print("[green]Recipes:[/green]")
        console.print(result)


def handle_get_recipe():
    """CLI flow for retrieving a single recipe by ID."""
    global qubicon_core
    if not qubicon_core:
        console.print("[red]Error: You must be logged in before getting a recipe.[/red]")
        return

    recipe_id = IntPrompt.ask("Enter the recipe ID")
    console.print(f"[cyan]Fetching recipe {recipe_id}...[/cyan]")
    recipe = qubicon_core.get_recipe(recipe_id)
    if recipe is None:
        console.print("[red]Failed to fetch recipe details.[/red]")
    else:
        console.print("[green]Recipe details:[/green]")
        console.print(recipe)


def handle_create_process_group():
    """CLI flow for creating a new process group."""
    global qubicon_core
    if not qubicon_core:
        console.print("[red]Error: You must be logged in before creating a process group.[/red]")
        return

    group_name = Prompt.ask("Enter a name for the new process group")
    status = Prompt.ask("Enter status for the group (default=RELEASED)", default="RELEASED")

    console.print(f"[cyan]Creating process group '{group_name}'...[/cyan]")
    try:
        group = qubicon_core.create_process_group(name=group_name, status=status)
        if group:
            console.print("[green]Process group created successfully![/green]")
            console.print(group)
        else:
            console.print("[red]Failed to create process group.[/red]")
    except Exception as e:
        console.print(f"[red]Error creating process group: {e}[/red]")


def handle_add_processes_to_group():
    """CLI flow for adding processes to an existing group."""
    global qubicon_core
    if not qubicon_core:
        console.print("[red]Error: You must be logged in before modifying a process group.[/red]")
        return

    group_id = IntPrompt.ask("Enter the group ID")
    console.print("Enter process IDs to add to group, comma-separated.")
    pids_str = Prompt.ask("Process IDs")
    try:
        process_ids = [int(pid.strip()) for pid in pids_str.split(",")]
    except ValueError:
        console.print("[red]Invalid input. Please enter numeric IDs, comma-separated.[/red]")
        return

    console.print(f"[cyan]Adding processes {process_ids} to group {group_id}...[/cyan]")
    try:
        updated_group = qubicon_core.add_processes_to_group(group_id, process_ids)
        if updated_group:
            console.print("[green]Processes added successfully![/green]")
            console.print(updated_group)
        else:
            console.print("[red]Failed to update process group.[/red]")
    except Exception as e:
        console.print(f"[red]Error updating process group: {e}[/red]")

def handle_create_process():
    """CLI flow for creating a new process from a chosen recipe."""
    global qubicon_core

    if not qubicon_core:
        console.print("[red]Error: You must be logged in before creating a process.[/red]")
        return

    # Ask the user for process details
    process_name = Prompt.ask("Enter a name for the new process")
    recipe_id = IntPrompt.ask("Enter the recipe ID to base this process on")

    # Optional: ask for description, or skip if not needed
    description = Prompt.ask("Enter process description (optional)", default="")

    # Create the process via QubiconCore
    console.print(f"[cyan]Creating process '{process_name}' using recipe ID {recipe_id}...[/cyan]")
    try:
        result = qubicon_core.create_process(
            name=process_name,
            recipe_id=recipe_id,
            description=description,
            # add other optional args if desired: online_equipment_ids, offline_equipment_ids, etc.
        )
        if result:
            console.print(f"[green]Process created successfully![/green]\n{result}")
        else:
            console.print("[red]Failed to create process.[/red]")
    except Exception as e:
        console.print(f"[red]Error creating process: {e}[/red]")

def fetch_process_channels():
    """Fetch and display channels for a specific process."""
    process_id = Prompt.ask("Enter the Process ID", default="0")
    process_phase_id = Prompt.ask("Enter Process Phase ID (or leave blank)", default=UNSET)
    channel_name = Prompt.ask("Enter Channel Name (or leave blank)", default=UNSET)
    node_types_input = Prompt.ask("Enter Node Types (comma-separated, e.g., INPUT,OUTPUT) or leave blank", default=UNSET)
    sensor_type_ids_input = Prompt.ask("Enter Sensor Type IDs (comma-separated, e.g., 1,2,3) or leave blank", default=UNSET)
    physical_quantity_unit_id = Prompt.ask("Enter Physical Quantity Unit ID (or leave blank)", default=UNSET)

    ensure_logged_in()

    try:
        # Prepare the keyword arguments dynamically
        kwargs = {"process_id": int(process_id)}

        # Add filters to kwargs only if they are not UNSET
        if process_phase_id is not UNSET:
            kwargs["process_phase_id"] = int(process_phase_id)
        if channel_name is not UNSET:
            kwargs["name"] = channel_name
        if node_types_input is not UNSET:
            kwargs["node_types"] = [x.strip() for x in node_types_input.split(",") if x.strip()]
        if sensor_type_ids_input is not UNSET:
            kwargs["sensor_type_ids"] = [int(x.strip()) for x in sensor_type_ids_input.split(",") if x.strip()]
        if physical_quantity_unit_id is not UNSET:
            kwargs["physical_quantity_unit_id"] = int(physical_quantity_unit_id)

        # Fetch channels using the core function
        channels = qubicon_core.extract_channels(**kwargs)

        # Check if the response contains channels
        if isinstance(channels, dict) and "content" in channels:
            # Display the channels in a table
            table = Table(title=f"Channels for Process {process_id}", expand=True)
            table.add_column("ID", justify="right", style="cyan")
            table.add_column("Name", justify="left", style="magenta")
            table.add_column("Node Type", justify="left", style="green")
            table.add_column("Unit", justify="left", style="blue")
            table.add_column("Physical Quantity", justify="left", style="yellow")
            table.add_column("Equipment Name", justify="left", style="white")

            for channel in channels["content"]:
                physical_quantity = channel.get("physicalQuantityUnit", {})
                table.add_row(
                    str(channel.get("id", "N/A")),
                    channel.get("name", "N/A"),
                    channel.get("nodeType", "N/A"),
                    physical_quantity.get("unit", "N/A"),
                    physical_quantity.get("name", "N/A"),
                    channel.get("equipmentName", "N/A"),
                )

            console.print(table)

        elif isinstance(channels, dict) and "error" in channels:
            console.print(f"[red]Error: {channels['error']}[/red]")
        else:
            console.print("[yellow]No channels found for the given process ID.[/yellow]")

    except Exception as e:
        console.print(f"[red]Error fetching process channels: {e}[/red]")

def display_models():
    """Fetch and display all computable models in a structured table format."""
    ensure_logged_in()
    try:
        # Fetch the models from the core function
        response = qubicon_core.list_computable_models()

        if "content" in response and isinstance(response["content"], list):
            # Create a Rich table for formatted output
            table = Table(title="Computable Models", expand=True)
            table.add_column("ID", justify="center", style="cyan", no_wrap=True)
            table.add_column("KPI Name", justify="center", style="magenta", no_wrap=True)
            table.add_column("Status", justify="center", style="green", no_wrap=True)
            table.add_column("Calculation Style", justify="center", style="yellow", no_wrap=True)
            table.add_column("Engine Type", justify="center", style="red", no_wrap=True)
            table.add_column("Update Date", justify="center", style="white", no_wrap=True)
            table.add_column("Creation Date", justify="center", style="white", no_wrap=True)

            # Add rows to the table
            for model in response["content"]:
                table.add_row(
                    str(model.get("id", "N/A")),
                    model.get("kpiName", "N/A"),
                    model.get("status", "N/A"),
                    model.get("calculationStyle", "N/A"),
                    model.get("engineType", "N/A"),
                    str(model.get("updateDate", "N/A")),
                    str(model.get("creationDate", "N/A")),
                )

            # Print the table
            console.print(table)
        else:
            console.print("[yellow]No models found or unexpected response format.[/yellow]")

    except Exception as e:
        console.print(f"[red]Error listing models: {e}[/red]")

def fetch_model_details():
    """Fetch and display details for a specific model."""
    model_id = Prompt.ask("Enter the model ID", default="0")
    ensure_logged_in()
    try:
        details = qubicon_core.fetch_model_details(int(model_id))
        if details:
            table = Table(title="Model Details", expand=True)
            table.add_column("Property", justify="right", style="cyan")
            table.add_column("Value", justify="left", style="magenta")

            for key, value in sorted(details.items()):
                # Format nested JSON-like objects
                if isinstance(value, (list, dict)):
                    formatted_value = json.dumps(value, indent=4, ensure_ascii=False)
                    table.add_row(str(key), formatted_value)
                else:
                    table.add_row(str(key), str(value))
            
            console.print(table)
        else:
            console.print("[red]No details found for the specified model.[/red]")
    except Exception as e:
        console.print(f"[red]Error fetching model details: {e}[/red]")

def list_physical_quantities():
    """Fetch and display all physical quantities with their units."""
    ensure_logged_in()
    try:
        # Fetch the physical quantities from the core function
        quantities = qubicon_core.list_physical_quantities()
        if quantities:
            # Create a Rich table for formatted output
            table = Table(title="Physical Quantities", expand=True)
            table.add_column("ID", justify="right", style="cyan")
            table.add_column("Name", justify="left", style="magenta")
            table.add_column("Units", justify="left", style="green")

            # Add rows to the table
            for quantity in sorted(quantities, key=lambda x: x['id']):
                # Extract units and format as a comma-separated list
                units = (
                    ", ".join(unit.get("unit", "N/A") for unit in quantity.get("units", []))
                    if "units" in quantity and isinstance(quantity["units"], list)
                    else "N/A"
                )
                table.add_row(
                    str(quantity["id"]),
                    quantity.get("name", "N/A"),
                    units,
                )

            # Print the table
            console.print(table)
        else:
            console.print("[red]No physical quantities found.[/red]")
    except Exception as e:
        console.print(f"[red]Error listing physical quantities: {e}[/red]")

def list_processes():
    """Fetch and display all processes."""
    ensure_logged_in()
    try:
        # Fetch the processes from the core function
        response = qubicon_core.list_processes()

        if "content" in response and isinstance(response["content"], list):
            processes = response["content"]

            if processes:
                # Create a table for displaying processes
                table = Table(title="Processes", expand=True)
                table.add_column("ID", justify="right", style="cyan")
                table.add_column("Name", justify="left", style="magenta")
                table.add_column("Description", justify="left", style="green")
                table.add_column("Status", justify="left", style="yellow")
                table.add_column("Type", justify="left", style="blue")
                table.add_column("Creation Date", justify="left", style="white")
                table.add_column("End Date", justify="left", style="white")

                for process in sorted(processes, key=lambda x: x['id']):
                    # Format dates from Unix timestamps
                    creation_date = process.get('creationDate', None)
                    end_date = process.get('endDate', None)
                    formatted_creation_date = (
                        datetime.utcfromtimestamp(creation_date / 1000).strftime('%Y-%m-%d %H:%M:%S')
                        if creation_date else "N/A"
                    )
                    formatted_end_date = (
                        datetime.utcfromtimestamp(end_date / 1000).strftime('%Y-%m-%d %H:%M:%S')
                        if end_date else "N/A"
                    )

                    # Add rows to the table
                    table.add_row(
                        str(process['id']),
                        process.get('name', 'N/A'),
                        process.get('description', 'N/A'),
                        process.get('status', 'N/A'),
                        process.get('type', 'N/A'),
                        formatted_creation_date,
                        formatted_end_date
                    )

                console.print(table)
            else:
                console.print("[yellow]No processes found.[/yellow]")
        elif "error" in response:
            console.print(f"[red]{response['error']}[/red]")
        else:
            console.print("[yellow]Unexpected response format.[/yellow]")

    except Exception as e:
        console.print(f"[red]Error listing processes: {e}[/red]")

def handle_process_data_extraction():
    """Interactive process data extraction."""
    ensure_logged_in()
    try:
        # Prompt for Process ID and validate it
        process_id = int(Prompt.ask("Enter the Process ID to extract data from", default="1"))

        # Fetch the start and end time for the process
        start_time = int(Prompt.ask("Enter the start time (Unix timestamp in ms)", default="0"))
        end_time = int(Prompt.ask("Enter the end time (Unix timestamp in ms)", default="0"))

        # Optional parameters for data extraction
        raw_mode = Confirm.ask("Enable raw mode?", default=False)
        export_rate = int(Prompt.ask("Enter the export rate (granularity in ms)", default="148"))
        at_sampling_timestamps_mode = Confirm.ask("Export at sampling timestamps?", default=False)
        included_report = Confirm.ask("Include report in export?", default=False)

        # Prompt for optional event IDs
        start_event_id_input = Prompt.ask("Enter start event ID (or leave blank)", default="")
        end_event_id_input = Prompt.ask("Enter end event ID (or leave blank)", default="")
        start_event_id = int(start_event_id_input) if start_event_id_input else UNSET
        end_event_id = int(end_event_id_input) if end_event_id_input else UNSET

        # Output format selection
        output_format = Prompt.ask("Enter your format of choice (csv or json)", choices=["csv", "json"], default="csv")
        output_file = Prompt.ask("Enter the output file path", default="process_data.csv" if output_format == "csv" else "process_data.json")

        # Extract process data using the core function
        response = qubicon_core.extract_process_data(
            process_id=process_id,
            start_time=start_time,
            end_time=end_time,
            raw_mode=raw_mode,
            export_rate=export_rate,
            at_sampling_timestamps_mode=at_sampling_timestamps_mode,
            included_report=included_report,
            start_event_id=start_event_id,
            end_event_id=end_event_id,
        )

        # Handle the response and save the output
        if "error" not in response:
            if output_format == "csv":
                # Convert JSON data to CSV and save
                df = pd.DataFrame(response["data"])  # Adjust based on the response structure
                df.to_csv(output_file, index=False)
                console.print(f"[green]Process data successfully saved to {output_file} in CSV format.[/green]")
            else:
                # Save as JSON
                with open(output_file, "w") as f:
                    json.dump(response, f, indent=4)
                console.print(f"[green]Process data successfully saved to {output_file} in JSON format.[/green]")
        else:
            console.print(f"[red]Error: {response['error']}[/red]")

    except Exception as e:
        console.print(f"[red]Unexpected error: {str(e)}[/red]")

def handle_logout():
    """Log out the user and clear session data."""
    global client, qubicon_core
    client = None
    qubicon_core = None
    console.print("[green]Logged out successfully.[/green]")

def export_model_in_importable_format():
    """Extract a model in a format that can be imported."""
    ensure_logged_in()
    model_id = Prompt.ask("Enter the model ID", default="0")
    output_file = Prompt.ask("Enter the output file path", default="exported_model.json")
    try:
        response = qubicon_core.export_model_to_json(int(model_id), output_file)
    except Exception as e:
        console.print(f"[red]Error extracting model: {e}[red]")

def create_computable_model():
    """Create a new computable model interactively."""
    ensure_logged_in()
    try:
        model_data = Prompt.ask("Enter the path to the JSON file")
        # turn json file into dict
        with open(model_data) as f:
            model_data = json.load(f)
        print(model_data)
        response = qubicon_core.create_computable_model(model_data)
        
        if response:
            console.print("[green]Model created successfully![green]")
    except Exception as e:
        console.print(f"[red]Error creating model: {e}[red]")

def main():
    """Entry point for the CLI application."""
    parser = argparse.ArgumentParser(description="Qubicon API Client")
    parser.add_argument("--interactive", action="store_true", help="Launch interactive mode")
    args = parser.parse_args()

    if args.interactive:
        interactive_mode()
    else:
        console.print("[yellow]Use --interactive to start the CLI.[/yellow]")

if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.abstract_import_file_mapping_dto_type import AbstractImportFileMappingDtoType
from qubicon.api.types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.any_import_file_mapping_column_dto import AnyImportFileMappingColumnDto


T = TypeVar("T", bound="OsmometerImportFileMappingDto")


@_attrs_define
class OsmometerImportFileMappingDto:
    """
    Attributes:
        type (Union[Unset, AbstractImportFileMappingDtoType]):
        columns (Union[Unset, List['AnyImportFileMappingColumnDto']]):
        sample_id_column (Union[Unset, int]):
        data_row (Union[Unset, int]):
    """

    type: Union[Unset, AbstractImportFileMappingDtoType] = UNSET
    columns: Union[Unset, List["AnyImportFileMappingColumnDto"]] = UNSET
    sample_id_column: Union[Unset, int] = UNSET
    data_row: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type: Union[Unset, str] = UNSET
        if not isinstance(self.type, Unset):
            type = self.type.value

        columns: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.columns, Unset):
            columns = []
            for columns_item_data in self.columns:
                columns_item = columns_item_data.to_dict()
                columns.append(columns_item)

        sample_id_column = self.sample_id_column

        data_row = self.data_row

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if type is not UNSET:
            field_dict["type"] = type
        if columns is not UNSET:
            field_dict["columns"] = columns
        if sample_id_column is not UNSET:
            field_dict["sampleIdColumn"] = sample_id_column
        if data_row is not UNSET:
            field_dict["dataRow"] = data_row

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.any_import_file_mapping_column_dto import AnyImportFileMappingColumnDto

        d = src_dict.copy()
        _type = d.pop("type", UNSET)
        type: Union[Unset, AbstractImportFileMappingDtoType]
        if isinstance(_type, Unset):
            type = UNSET
        else:
            type = AbstractImportFileMappingDtoType(_type)

        columns = []
        _columns = d.pop("columns", UNSET)
        for columns_item_data in _columns or []:
            columns_item = AnyImportFileMappingColumnDto.from_dict(columns_item_data)

            columns.append(columns_item)

        sample_id_column = d.pop("sampleIdColumn", UNSET)

        data_row = d.pop("dataRow", UNSET)

        osmometer_import_file_mapping_dto = cls(
            type=type,
            columns=columns,
            sample_id_column=sample_id_column,
            data_row=data_row,
        )

        osmometer_import_file_mapping_dto.additional_properties = d
        return osmometer_import_file_mapping_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class GetOnlineEquipmentsTypesItem(str, Enum):
    EQUIPMENT = "EQUIPMENT"
    EQUIPMENT_ROOT = "EQUIPMENT_ROOT"
    FOLDER = "FOLDER"

    def __str__(self) -> str:
        return str(self.value)



from qubicon import QubiconCore

# Mock client with necessary attributes
class MockClient:
    def __init__(self):
        self.base_url = "https://master.qub-lab.io"
        self.token = None

mock_client = MockClient()
core = QubiconCore(client=mock_client)

print("\n--- Testing login_user ---")
try:
    token = core.login_user(username="qubiconClient", password="qubiconClient1!")
    print("Token:", token)
except Exception as e:
    print(f"login_user failed: {e}")




"""Contains all the data models used in inputs/outputs"""

from .abstract_computable_model_dto import AbstractComputableModelDto
from .abstract_computable_model_dto_calculation_style import AbstractComputableModelDtoCalculationStyle
from .abstract_computable_model_dto_engine_type import AbstractComputableModelDtoEngineType
from .abstract_computable_model_dto_status import AbstractComputableModelDtoStatus
from .abstract_import_file_mapping_dto import AbstractImportFileMappingDto
from .abstract_import_file_mapping_dto_type import AbstractImportFileMappingDtoType
from .abstract_import_file_progress import AbstractImportFileProgress
from .abstract_import_file_progress_type import AbstractImportFileProgressType
from .abstract_job_result import AbstractJobResult
from .abstract_job_result_job_type import AbstractJobResultJobType
from .any_import_file_mapping_column_dto import AnyImportFileMappingColumnDto
from .authentication_dto import AuthenticationDto
from .authentication_token_dto import AuthenticationTokenDto
from .ce_dex_import_file_mapping_dto import CeDexImportFileMappingDto
from .channel_data_channel_request_key_dto import ChannelDataChannelRequestKeyDto
from .channel_data_channel_request_key_dto_type import ChannelDataChannelRequestKeyDtoType
from .channel_data_dto import ChannelDataDto
from .channel_data_key_dto import ChannelDataKeyDto
from .channel_data_key_dto_type import ChannelDataKeyDtoType
from .channel_data_pair_dto import ChannelDataPairDto
from .channel_dto import ChannelDto
from .channel_dto_data_presentation_type import ChannelDtoDataPresentationType
from .channel_dto_node_type import ChannelDtoNodeType
from .channel_dto_type import ChannelDtoType
from .computable_model_input_dto import ComputableModelInputDto
from .computable_model_output_dto import ComputableModelOutputDto
from .direct_to_process_import_file_progress import DirectToProcessImportFileProgress
from .direct_to_process_import_file_progress_type import DirectToProcessImportFileProgressType
from .external_python_computable_model_dto import ExternalPythonComputableModelDto
from .get_channels_node_types_item import GetChannelsNodeTypesItem
from .get_equipment_variable_nodes_node_usage_types_item import GetEquipmentVariableNodesNodeUsageTypesItem
from .get_list_of_computable_models_calculation_styles_item import GetListOfComputableModelsCalculationStylesItem
from .get_list_of_computable_models_statuses_item import GetListOfComputableModelsStatusesItem
from .get_list_of_computable_models_types_item import GetListOfComputableModelsTypesItem
from .get_offline_equipment_statuses_item import GetOfflineEquipmentStatusesItem
from .get_online_equipments_statuses_item import GetOnlineEquipmentsStatusesItem
from .get_online_equipments_types_item import GetOnlineEquipmentsTypesItem
from .get_process_groups_status import GetProcessGroupsStatus
from .get_processes_statuses_item import GetProcessesStatusesItem
from .get_processes_types_item import GetProcessesTypesItem
from .gmp_track_confirmation_dto import GmpTrackConfirmationDto
from .gmp_track_confirmation_dto_role import GmpTrackConfirmationDtoRole
from .gmp_track_dto import GmpTrackDto
from .import_offline_equipment_data_file_async_body import ImportOfflineEquipmentDataFileAsyncBody
from .job_dto import JobDto
from .job_dto_status import JobDtoStatus
from .job_dto_type import JobDtoType
from .job_process_offline_import_result import JobProcessOfflineImportResult
from .job_sampling_offline_data_import_result import JobSamplingOfflineDataImportResult
from .job_sampling_offline_import_auto_result import JobSamplingOfflineImportAutoResult
from .jython_computable_model_dto import JythonComputableModelDto
from .minimum_process_dto import MinimumProcessDto
from .minimum_process_dto_python_engine_status import MinimumProcessDtoPythonEngineStatus
from .minimum_process_dto_status import MinimumProcessDtoStatus
from .minimum_process_group_dto import MinimumProcessGroupDto
from .minimum_process_group_dto_status import MinimumProcessGroupDtoStatus
from .multiplex_chart_data_channels_request_dto import MultiplexChartDataChannelsRequestDto
from .multiplex_chart_data_dto import MultiplexChartDataDto
from .nano_drop_import_file_mapping_dto import NanoDropImportFileMappingDto
from .offline_equipment_dto import OfflineEquipmentDto
from .offline_equipment_dto_status import OfflineEquipmentDtoStatus
from .offline_equipment_dto_type import OfflineEquipmentDtoType
from .online_equipment_dto import OnlineEquipmentDto
from .online_equipment_dto_equipment_data_refresh_status import OnlineEquipmentDtoEquipmentDataRefreshStatus
from .online_equipment_dto_equipment_metadata_rescan_status import OnlineEquipmentDtoEquipmentMetadataRescanStatus
from .online_equipment_dto_status import OnlineEquipmentDtoStatus
from .online_equipment_dto_type import OnlineEquipmentDtoType
from .online_equipment_node_object_dto import OnlineEquipmentNodeObjectDto
from .online_equipment_node_object_dto_equipment_type import OnlineEquipmentNodeObjectDtoEquipmentType
from .online_equipment_node_object_dto_node_opc_ua_type import OnlineEquipmentNodeObjectDtoNodeOpcUaType
from .online_equipment_node_variable_dto import OnlineEquipmentNodeVariableDto
from .online_equipment_node_variable_dto_equipment_type import OnlineEquipmentNodeVariableDtoEquipmentType
from .online_equipment_node_variable_dto_node_opc_ua_type import OnlineEquipmentNodeVariableDtoNodeOpcUaType
from .online_equipment_node_variable_dto_signal_type import OnlineEquipmentNodeVariableDtoSignalType
from .osmometer_import_file_mapping_dto import OsmometerImportFileMappingDto
from .page_dto_abstract_computable_model_dto import PageDtoAbstractComputableModelDto
from .page_dto_minimum_process_group_dto import PageDtoMinimumProcessGroupDto
from .page_dto_offline_equipment_dto import PageDtoOfflineEquipmentDto
from .page_dto_online_equipment_dto import PageDtoOnlineEquipmentDto
from .page_dto_online_equipment_node_variable_dto import PageDtoOnlineEquipmentNodeVariableDto
from .page_dto_process_dto import PageDtoProcessDto
from .page_meta_data_dto import PageMetaDataDto
from .ph_meter_import_file_mapping_dto import PHMeterImportFileMappingDto
from .photo_meter_import_file_mapping_dto import PhotoMeterImportFileMappingDto
from .physical_quantity_dto import PhysicalQuantityDto
from .physical_quantity_dto_status import PhysicalQuantityDtoStatus
from .physical_quantity_unit_dto import PhysicalQuantityUnitDto
from .physical_quantity_unit_dto_status import PhysicalQuantityUnitDtoStatus
from .privilege_dto import PrivilegeDto
from .process_dto import ProcessDto
from .process_dto_process_archive_transition_status import ProcessDtoProcessArchiveTransitionStatus
from .process_dto_python_engine_status import ProcessDtoPythonEngineStatus
from .process_dto_status import ProcessDtoStatus
from .process_dto_type import ProcessDtoType
from .process_phase_dto import ProcessPhaseDto
from .process_phase_dto_status import ProcessPhaseDtoStatus
from .process_phase_dto_type import ProcessPhaseDtoType
from .public_process_export_request_dto import PublicProcessExportRequestDto
from .publish_schema_dto import PublishSchemaDto
from .publish_schema_dto_db_column_type import PublishSchemaDtoDbColumnType
from .publish_schema_dto_field_data_type import PublishSchemaDtoFieldDataType
from .refresh_token_dto import RefreshTokenDto
from .role_dto import RoleDto
from .role_dto_name import RoleDtoName
from .sensor_type_dto import SensorTypeDto
from .server_sent_event_abstract_server_sent_event_dto import ServerSentEventAbstractServerSentEventDto
from .stored_file_dto import StoredFileDto
from .stream_dto import StreamDto
from .stream_events_types_item import StreamEventsTypesItem
from .user_dto import UserDto
from .user_dto_login_type import UserDtoLoginType
from .user_dto_user_deactivation_reason import UserDtoUserDeactivationReason
from .user_editor_dto import UserEditorDto
from .user_login_dto import UserLoginDto
from .user_login_dto_login_type import UserLoginDtoLoginType
from .vi_cell_import_file_mapping_dto import ViCellImportFileMappingDto
from .xlsx_parsing_error_model import XlsxParsingErrorModel
from .xlsx_parsing_error_model_throwable import XlsxParsingErrorModelThrowable
from .xlsx_parsing_error_model_throwable_stack_trace_item import XlsxParsingErrorModelThrowableStackTraceItem
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlags,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item_annotations_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemAnnotationsItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_defined_packages_item_declared_annotations_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemDeclaredAnnotationsItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParent,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent_defined_packages_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentDefinedPackagesItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent_defined_packages_item_annotations_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentDefinedPackagesItemAnnotationsItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent_defined_packages_item_declared_annotations_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentDefinedPackagesItemDeclaredAnnotationsItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent_unnamed_module import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModule,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent_unnamed_module_annotations_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModuleAnnotationsItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent_unnamed_module_declared_annotations_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModuleDeclaredAnnotationsItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent_unnamed_module_descriptor import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModuleDescriptor,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_parent_unnamed_module_layer import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModuleLayer,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_unnamed_module import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModule,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_unnamed_module_annotations_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModuleAnnotationsItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_unnamed_module_declared_annotations_item import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModuleDeclaredAnnotationsItem,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_unnamed_module_descriptor import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModuleDescriptor,
)
from .xlsx_parsing_error_model_throwable_stack_trace_item_include_info_flags_unnamed_module_layer import (
    XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModuleLayer,
)

__all__ = (
    "AbstractComputableModelDto",
    "AbstractComputableModelDtoCalculationStyle",
    "AbstractComputableModelDtoEngineType",
    "AbstractComputableModelDtoStatus",
    "AbstractImportFileMappingDto",
    "AbstractImportFileMappingDtoType",
    "AbstractImportFileProgress",
    "AbstractImportFileProgressType",
    "AbstractJobResult",
    "AbstractJobResultJobType",
    "AnyImportFileMappingColumnDto",
    "AuthenticationDto",
    "AuthenticationTokenDto",
    "CeDexImportFileMappingDto",
    "ChannelDataChannelRequestKeyDto",
    "ChannelDataChannelRequestKeyDtoType",
    "ChannelDataDto",
    "ChannelDataKeyDto",
    "ChannelDataKeyDtoType",
    "ChannelDataPairDto",
    "ChannelDto",
    "ChannelDtoDataPresentationType",
    "ChannelDtoNodeType",
    "ChannelDtoType",
    "ComputableModelInputDto",
    "ComputableModelOutputDto",
    "DirectToProcessImportFileProgress",
    "DirectToProcessImportFileProgressType",
    "ExternalPythonComputableModelDto",
    "GetChannelsNodeTypesItem",
    "GetEquipmentVariableNodesNodeUsageTypesItem",
    "GetListOfComputableModelsCalculationStylesItem",
    "GetListOfComputableModelsStatusesItem",
    "GetListOfComputableModelsTypesItem",
    "GetOfflineEquipmentStatusesItem",
    "GetOnlineEquipmentsStatusesItem",
    "GetOnlineEquipmentsTypesItem",
    "GetProcessesStatusesItem",
    "GetProcessesTypesItem",
    "GetProcessGroupsStatus",
    "GmpTrackConfirmationDto",
    "GmpTrackConfirmationDtoRole",
    "GmpTrackDto",
    "ImportOfflineEquipmentDataFileAsyncBody",
    "JobDto",
    "JobDtoStatus",
    "JobDtoType",
    "JobProcessOfflineImportResult",
    "JobSamplingOfflineDataImportResult",
    "JobSamplingOfflineImportAutoResult",
    "JythonComputableModelDto",
    "MinimumProcessDto",
    "MinimumProcessDtoPythonEngineStatus",
    "MinimumProcessDtoStatus",
    "MinimumProcessGroupDto",
    "MinimumProcessGroupDtoStatus",
    "MultiplexChartDataChannelsRequestDto",
    "MultiplexChartDataDto",
    "NanoDropImportFileMappingDto",
    "OfflineEquipmentDto",
    "OfflineEquipmentDtoStatus",
    "OfflineEquipmentDtoType",
    "OnlineEquipmentDto",
    "OnlineEquipmentDtoEquipmentDataRefreshStatus",
    "OnlineEquipmentDtoEquipmentMetadataRescanStatus",
    "OnlineEquipmentDtoStatus",
    "OnlineEquipmentDtoType",
    "OnlineEquipmentNodeObjectDto",
    "OnlineEquipmentNodeObjectDtoEquipmentType",
    "OnlineEquipmentNodeObjectDtoNodeOpcUaType",
    "OnlineEquipmentNodeVariableDto",
    "OnlineEquipmentNodeVariableDtoEquipmentType",
    "OnlineEquipmentNodeVariableDtoNodeOpcUaType",
    "OnlineEquipmentNodeVariableDtoSignalType",
    "OsmometerImportFileMappingDto",
    "PageDtoAbstractComputableModelDto",
    "PageDtoMinimumProcessGroupDto",
    "PageDtoOfflineEquipmentDto",
    "PageDtoOnlineEquipmentDto",
    "PageDtoOnlineEquipmentNodeVariableDto",
    "PageDtoProcessDto",
    "PageMetaDataDto",
    "PHMeterImportFileMappingDto",
    "PhotoMeterImportFileMappingDto",
    "PhysicalQuantityDto",
    "PhysicalQuantityDtoStatus",
    "PhysicalQuantityUnitDto",
    "PhysicalQuantityUnitDtoStatus",
    "PrivilegeDto",
    "ProcessDto",
    "ProcessDtoProcessArchiveTransitionStatus",
    "ProcessDtoPythonEngineStatus",
    "ProcessDtoStatus",
    "ProcessDtoType",
    "ProcessPhaseDto",
    "ProcessPhaseDtoStatus",
    "ProcessPhaseDtoType",
    "PublicProcessExportRequestDto",
    "PublishSchemaDto",
    "PublishSchemaDtoDbColumnType",
    "PublishSchemaDtoFieldDataType",
    "RefreshTokenDto",
    "RoleDto",
    "RoleDtoName",
    "SensorTypeDto",
    "ServerSentEventAbstractServerSentEventDto",
    "StoredFileDto",
    "StreamDto",
    "StreamEventsTypesItem",
    "UserDto",
    "UserDtoLoginType",
    "UserDtoUserDeactivationReason",
    "UserEditorDto",
    "UserLoginDto",
    "UserLoginDtoLoginType",
    "ViCellImportFileMappingDto",
    "XlsxParsingErrorModel",
    "XlsxParsingErrorModelThrowable",
    "XlsxParsingErrorModelThrowableStackTraceItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlags",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemAnnotationsItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsDefinedPackagesItemDeclaredAnnotationsItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParent",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentDefinedPackagesItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentDefinedPackagesItemAnnotationsItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentDefinedPackagesItemDeclaredAnnotationsItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModule",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModuleAnnotationsItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModuleDeclaredAnnotationsItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModuleDescriptor",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsParentUnnamedModuleLayer",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModule",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModuleAnnotationsItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModuleDeclaredAnnotationsItem",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModuleDescriptor",
    "XlsxParsingErrorModelThrowableStackTraceItemIncludeInfoFlagsUnnamedModuleLayer",
)
